#
# Collective Knowledge (dealing with table)
#
# See CK LICENSE.txt for licensing details
# See CK COPYRIGHT.txt for copyright details
#
# Developer: Grigori Fursin, Grigori.Fursin@cTuning.org, http://fursin.net
#

cfg={}  # Will be updated by CK (meta description of this module)
work={} # Will be updated by CK (temporal data)
ck=None # Will be updated by CK (initialized CK kernel) 

# Local settings

##############################################################################
# Initialize module

def init(i):
    """

    Input:  {}

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """
    return {'return':0}

##############################################################################
# draw table

def draw(i):
    """
    Input:  {
              table   - table to draw [[],[],[]...], [[],[],[]...] ...]

              (out)   - txt (default) or html
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0

              string       - output
            }

    """

    o=i.get('out','')

    table=i.get('table',[])

    s=''

    if len(table)>0:
       lx=len(table[0])

       lwidth=[]
       for l in range(0, lx):
           lwidth.append(-1)

       # If 'txt', check length of all entries
       if o=='txt':
          for t in table:
              for l in range(0, lx):
                  sx=str(t[l])
                  lw=lwidth[l]
                  if lw==-1 or len(sx)>lw: 
                     lwidth[l]=len(sx)


          for t in table:
              for l in range(0, lx):
                  sx=str(t[l])
                  lw=lwidth[l]

                  s+=sx.ljust(lw+2)
              s+='\n'
       else:
          s='<html>\n'
          s+=' <body>\n'
          s+='  <table border="1">\n'
          for t in table:
              s+='  <tr>\n'
              for l in range(0, lx):
                  sx=str(t[l])
                  s+='    <td>'+sx+'</td>\n'
              s+='   </tr>\n'
          s+='  </table>\n'
          s+=' </body>\n'
          s+='<html>\n'

    return {'return':0, 'string':s}

##############################################################################
# prepare table (in HTML and LaTex)

def prepare(i):
    """
    Input:  {
              table
              table_header
              (table_custom)

              (table_style)
              (header_style)
              (header_element_style)
              (element_style)
              (row_style)

              (html_before_table)
              (html_after_table)

              (tex_before_table)
              (tex_after_table)

              (record_html) - file (with path) to record produced HTML
              (record_tex) - file (with path) to record produced TEX

            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0

              html - prepared HTML
              tex  - prepared :aTex
            }

    """

    table=i['table']
    table_header=i['table_header']
    table_custom=i.get('table_custom',{})

    h=i.get('html_before_table','') # HTML
    t=i.get('tex_before_table','')  # LaTex

    ts=i.get('table_style','')
    hs=i.get('header_style','')
    rs=i.get('row_style','')
    hes=i.get('header_element_style','')
    es=i.get('element_style','')

    # Preparing Header
    h+='<table '+ts+'>\n'

    t+='    \\begin{tabular}{|'

    for x in table_header:
        t+=x.get('tex','')+'|'

    t+='}\n'

    h+=' <tr '+hs+'>\n'
    t+='     \\hline\n'
    t+='      '

    first=True
    for x in table_header:
        n=x.get('name','')

        if first:
           first=False
        else:
           t+=' & '

        t+='\\textbf{'+n+'}'

        if x.get('html_change_space','')=='yes':
           n=n.replace(' ','&nbsp;')

        h+='  <td '+hes+'>\n'
        h+='   <b>'+n+'</b>\n'
        h+='  </td>\n'

    h+=' </tr>\n'
    t+=' \\\\ \n'

    # Preparing table
    for ix in range(0, len(table)):
        x=table[ix]

        cx={}
        if ix<len(table_custom):
           cx=table_custom[ix]

        rs1=cx.get('row_style','')

        h+=' <tr '+rs+' '+rs1+'>\n'

        t+='     \\hline\n'
        t+='      '

        first=True
        for iy in range(0, len(x)):
            st={}
            if iy<len(table_header):
               st=table_header[iy]

            y=cx.get('field_'+str(iy)+'_html','')
            if y=='':
               y=str(x[iy])

            if st.get('html_change_space','')=='yes':
               y=y.replace(' ','&nbsp;')

            y=y.replace('\\newline',' ')

            h+='  <td '+es+'>\n'
            if st.get('html_before','')!='':
               h+='   '+st['html_before']
            h+='   '+str(y)+'\n'
            if st.get('html_after','')!='':
               h+='   '+st['html_after']
            h+='  </td>\n'

            z=cx.get('field_'+str(iy)+'_tex','')
            if z=='':
               z=str(x[iy])

            if first:
               first=False
            else:
               t+=' & '

            if st.get('tex_before','')!='':
               t+=st['tex_before']
            t+=' '+str(z)+' '
            if st.get('tex_after','')!='':
               t+=st['tex_after']

        t+='\\\\\n'

        h+=' </tr>\n'
        
    t+="     \\hline\n"

    # Finalizing
    h+='</table>\n'
    h+=i.get('html_after_table','')

    t+='    \\end{tabular}'
    t+='    '+i.get('tex_after_table','')

    # Check if record
    if i.get('record_html','')!='':
       r=ck.save_text_file({'text_file':i['record_html'], 'string':h})

    if i.get('record_tex','')!='':
       r=ck.save_text_file({'text_file':i['record_tex'], 'string':t})

    return {'return':0, 'html':h, 'tex':t}



#
# Collective Knowledge (CK)
#
# See CK LICENSE.txt for licensing details
# See CK COPYRIGHT.txt for copyright details
#
# Developer: Grigori Fursin
#

import urllib
import json
import sys

try:
    from urllib.parse import urlencode
    from urllib.parse import quote_plus
except:
    from urllib import urlencode
    from urllib import quote_plus

try:
    import urllib.request as urllib2
except:
    import urllib2

##########################################################################


def request(i):
    """Web request to cKnowledge.org server
       Target audience: CK kernel and low-level developers

    Args:    
              get (dict): GET parameters
              post (dict): POST parameters

    Returns:
              (dict): Unified CK dictionary:

                return (int): return code =  0, if successful
                                          >  0, if error
                (error) (str): error text if return > 0

                string (str): returned string from the server
                dict (dict): JSON string converted to dict (if possible)

    """

    url = 'https://cKnowledge.org/api.php?'

    # Prepare dict to send to remote server
    ii = i.get('get', {})

    started = False
    for k in ii:
        v = ii[k]
        if started:
            url += '&'
        started = True
        url += k+'='+quote_plus(v)

    # Request
    request = urllib2.Request(url)

    # Connect
    try:
        f = urllib2.urlopen(request)
    except Exception as e:
        return {'return': 1, 'error': 'Access to cKnowledge.org failed ('+format(e)+')'}

    # Read
    try:
        s = f.read()
    except Exception as e:
        return {'return': 1, 'error': 'Failed to read stream from cKnowledge.org ('+format(e)+')'}

    # CLose
    try:
        f.close()
    except Exception as e:
        return {'return': 1, 'error': 'Failed to close stream from cKnowledge.org ('+format(e)+')'}

    # Check UTF
    try:
        s = s.decode('utf8')
    except Exception as e:
        pass

    # Check output
    d = {}

    # Try to convert output to dictionary
    try:
        d = json.loads(s)
    except Exception as e:
        pass

    return {'return': 0, 'string': s, 'dict': d}


##########################################################################
def access_ck_api(i):
    """Universal web request to the CK server (usually cKnowledge.io)
       Target audience: CK kernel and low-level developers

    Args:    
              url (str): URL API
              (dict) (dict): dict to send to above URL

    Returns:
              (dict): Unified CK dictionary:

                return (int): return code =  0, if successful
                                          >  0, if error
                (error) (str): error text if return > 0

                dict (dict): dictionary from the CK server

    """

    import ck.strings

    url = i['url']
    d = i.get('dict', {})

    # Import modules compatible with Python 2.x and 3.x
    import urllib

    try:
        import urllib.request as urllib2
    except:
        import urllib2

    try:
        from urllib.parse import urlencode
    except:
        from urllib import urlencode

    # Prepare post variables
    r = ck.strings.dump_json({'dict': d, 'skip_indent': 'yes'})
    if r['return'] > 0:
        return r

    s = r['string']
    if sys.version_info[0] > 2:
        s = s.encode('utf8')

    # Prepare request
    request = urllib2.Request(url, s, {'Content-Type': 'application/json'})

    # Connect
    try:
        f = urllib2.urlopen(request)
    except Exception as e:
        return {'return': 1, 'error': 'Access to the CK portal failed ('+format(e)+')'}

    # Read from Internet
    try:
        s = f.read()
        f.close()
    except Exception as e:
        return {'return': 1, 'error': 'Failed reading stream from the CK portal ('+format(e)+')'}

    # Check output
    try:
        s = s.decode('utf8')
    except Exception as e:
        pass

    # Try to convert output to dictionary
    r = ck.strings.convert_json_str_to_dict(
        {'str': s, 'skip_quote_replacement': 'yes'})
    if r['return'] > 0:
        return {'return': 1, 'error': 'can\'t parse output from the CK portal ('+r['error']+'):\n'+s[:256]+'\n\n...)'}

    d = r['dict']

    if 'return' in d:
        d['return'] = int(d['return'])
    else:
        return {'return': 99, 'error': 'repsonse doesn\'t follow the CK API standard'}

    return {'return': 0, 'dict': d}



#
# Collective Knowledge (CK web front-end)
#
# See CK LICENSE.txt for licensing details
# See CK COPYRIGHT.txt for copyright details
#
# Developer: cTuning foundation
#

cfg={}  # Will be updated by CK (meta description of this module)
work={} # Will be updated by CK (temporal data)
ck=None # Will be updated by CK (initialized CK kernel) 

# Local settings

import os

fdata_name='ck_top_data'
fmodule_name='ck_top_module'
frepo_name='ck_top_repo'

form_name='ck_top_form'
onchange='document.'+form_name+'.submit();'

##############################################################################
# Initialize module

def init(i):
    """

    Input:  {}

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """
    return {'return':0}


##############################################################################
# display webpage with interactive dashboard widget

def display(args):
    """
    Input:  {
              (host)        - Internal web server host
              (port)        - Internal web server port

              (wfe_host)    - External web server host
              (wfe_port)    - External web server port

              (extra_url)   - extra URL
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    # Filter and pass some arguments from command line to browser
    extra_url = { key:args[key] for key in args if key in ['scenario', 'global'] }
    extra_url = "&".join("{0}={1}".format(key, extra_url[key]) for key in extra_url)

    args['action'] = 'start'
    args['module_uoa'] = 'web'
    args['browser'] = 'yes'
    args['template'] = 'dashboard'
    args['cid'] = ''
    args['extra_url'] = extra_url

    return ck.access(args)



#
# Collective Knowledge (CK solution to automate the installation of program workflows and their benchmarking)
#
# See CK LICENSE.txt for licensing details
# See CK COPYRIGHT.txt for copyright details
#
# Developer: Grigori Fursin, Grigori.Fursin@cTuning.org, http://fursin.net
#

cfg={}  # Will be updated by CK (meta description of this module)
work={} # Will be updated by CK (temporal data)
ck=None # Will be updated by CK (initialized CK kernel) 

# Local settings

##############################################################################
# Initialize module

def init(i):
    """

    Input:  {}

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """
    return {'return':0}



#
# Collective Knowledge ($#desc#$)
#
# $#license#$
# $#copyright#$
#
# Developer: $#developer#$
#

cfg = {}  # Will be updated by CK (meta description of this module)
work = {}  # Will be updated by CK (temporal data)
ck = None  # Will be updated by CK (initialized CK kernel)

# Local settings

##############################################################################
# Initialize module


def init(i):
    """

    Input:  {}

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """
    return {'return': 0}



#
# Collective Knowledge: crowdbenchmarking using ARM's workload automation and CK
#
# See CK LICENSE.txt for licensing details
# See CK COPYRIGHT.txt for copyright details
#
# Developer: Grigori Fursin, Grigori.Fursin@cTuning.org, http://fursin.net
#

cfg={}  # Will be updated by CK (meta description of this module)
work={} # Will be updated by CK (temporal data)
ck=None # Will be updated by CK (initialized CK kernel) 

# Local settings
compiler_choices='#choices#compiler_flags#'

line='================================================================'

fsummary='summary.json'
fclassification='classification.json'
fgraph='tmp-reactions-graph.json'
ffstat='ck-stat-flat-characteristics.json'

form_name='wa_web_form'
onchange='document.'+form_name+'.submit();'

hextra=''
#hextra+='<i><center>\n'
#hextra+=' [ <a href="http://cKnowledge.org/ai">Community-driven AI R&D powered by CK</a> ], '
#hextra+=' [ <i><a href="http://dividiti.blogspot.fr/2017/02/we-received-test-of-time-award-for-our.html">CGO\'17 test of time award for our interdisiplinary R&D</a></i> ], '
#hextra+=' [ <b><a href="http://cKnowledge.org/android-apps.html">Android app to crowd-optimize DNN engines and models</a></b> ], '
#hextra+=' [ <a href="https://github.com/ctuning/ck-caffe2">CK-Caffe2 GitHub</a> / <a href="https://github.com/dividiti/ck-caffe">CK-Caffe GitHub</a> ], '
#hextra+=' [ <a href="https://github.com/ctuning/ck-tensorflow">CK-TensorFlow GitHub</a> ], '
#hextra+=' [ <a href="https://en.wikipedia.org/wiki/Collective_Knowledge_(software)">Wikipedia</a>, \n'
#hextra+='<a href="https://www.researchgate.net/publication/304010295_Collective_Knowledge_Towards_RD_Sustainability">paper 1</a>, \n'
#hextra+='<a href="https://arxiv.org/abs/1506.06256">Paper 2</a>, \n'
#hextra+='<a href="https://www.youtube.com/watch?v=Q94yWxXUMP0">YouTube CK intro</a> ] \n'
#hextra+='</center></i>\n'
#hextra+='<br>\n'

selector=[{'name':'Scenario', 'key':'crowd_uid', 'module_uoa':'65477d547a49dd2c', 'module_key':'##dict#title'},
          {'name':'DNN engine', 'key':'engine'},
          {'name':'Model', 'key':'model'},
          {'name':'Platform', 'key':'plat_name','new_line':'no'},
          {'name':'OS', 'key':'os_name'},
          {'name':'CPU', 'key':'cpu_name', 'new_line':'no'},
          {'name':'CPU ABI', 'key':'cpu_abi'},
          {'name':'GPU', 'key':'gpu_name'}]

abis=[{'abi':'arm64-v8a',
       'os_uoa':'android21-arm64'},
      {'abi':'armeabi-v7a',
       'os_uoa':'android21-arm-v7a'}]

libxopenme='librtlxopenme.so'

##############################################################################
# Initialize module

def init(i):
    """

    Input:  {}

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """
    return {'return':0}

##############################################################################
# show results

def show(i):
    """
    Input:  {
               (crowd_module_uoa) - if rendered from experiment crowdsourcing
               (crowd_key)        - add extra name to Web keys to avoid overlapping with original crowdsourcing HTML
               (crowd_on_change)  - reuse onchange doc from original crowdsourcing HTML

               (highlight_behavior_uid) - highlight specific result (behavior)!
               (highlight_by_user)      - highlight all results from a given user

               (all)     - if 'yes', view all results

               (minimal) - if 'yes', use in interactive papers
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    import os
    import copy
    import time

    st=''

    minimal=i.get('minimal','')=='yes'

    cmuoa=i.get('crowd_module_uoa','')
    ckey=i.get('crowd_key','')

    debug=(i.get('debug','')=='yes')
#    debug=True

    conc=i.get('crowd_on_change','')
    if conc=='':
        conc=onchange

    hi_uid=i.get('highlight_behavior_uid','')
    hi_user=i.get('highlight_by_user','')

#    h='<hr>\n'
    h='<center>\n'
    h+='\n\n<script language="JavaScript">function copyToClipboard (text) {window.prompt ("Copy to clipboard: Ctrl+C, Enter", text);}</script>\n\n' 

    h+=hextra

#    h+='<hr>\n'
    h+='<br>\n'

    # Check host URL prefix and default module/action
    rx=ck.access({'action':'form_url_prefix',
                  'module_uoa':'wfe',
                  'host':i.get('host',''), 
                  'port':i.get('port',''), 
                  'template':i.get('template','')})
    if rx['return']>0: return rx
    url0=rx['url']
    template=rx['template']

    url=url0
    action=i.get('action','')
    muoa=i.get('module_uoa','')

    st=''

    url+='action=index&module_uoa=wfe&native_action='+action+'&'+'native_module_uoa='+muoa
    url1=url

    # List entries
    dt=time.time()
    ii={'action':'search',
        'module_uoa':work['self_module_uid'],
        'add_meta':'yes'}

    if cmuoa!='':
        ii['module_uoa']=cmuoa

    r=ck.access(ii)
    if r['return']>0: return r

    if debug: h+='\n<p>Debug time (CK query): '+str(time.time()-dt)+' sec.<p>\n'

    lst=r['lst']

    # Find unique variables
    dt=time.time()

    choices={}
    mchoices={} # cache of UID -> alias choices
    wchoices={}

    cache_meta={}

    for q in lst:
        d=q['meta']
        meta=d.get('meta',{})

        # Process some derivatives
        scenario=meta.get('crowd_uid','')

        kscenario=cfg['module_deps']['experiment.scenario.mobile']+':'+scenario
        if kscenario not in cache_meta:
            r=ck.access({'action':'load',
                         'module_uoa':cfg['module_deps']['experiment.scenario.mobile'],
                         'data_uoa':scenario})
            if r['return']>0: return r
            xd=r['dict']

            # Find model size
            for q in xd.get('files',[]):
                if q.get('model_weights','')=='yes':
                    xd['model_weights_size']=int(q.get('file_size',0)/1E6)
                    break

            cache_meta[kscenario]=xd
        else:
            xd=cache_meta[kscenario]

        if meta.get('engine','')=='': meta['engine']=xd.get('engine','')
        if meta.get('model','')=='': meta['model']=xd.get('model','')

        meta['engine_meta']=xd.get('engine_meta',{})

        # Process selector meta
        for kk in selector:
            kx=kk['key']
            k=ckey+kx

            if k not in choices: 
                choices[k]=[]
                wchoices[k]=[{'name':'','value':''}]

            v=meta.get(kx,'')
            if v!='':

                if v not in choices[k]: 
                    choices[k].append(v)

                    muoa=kk.get('module_uoa','')
                    vv=v
                    if muoa!='':
                        if k not in mchoices:
                            mchoices[k]={}

                        vv=mchoices[k].get(v,'')
                        if vv=='':
                            r=ck.access({'action':'load',
                                         'module_uoa':muoa,
                                         'data_uoa':v})
                            if r['return']==0:
                                mk=kk.get('module_key','')
                                if mk=='': mk='##data_name'

                                rx=ck.get_by_flat_key({'dict':r, 'key':mk})
                                if rx['return']>0: return rx
                                vv=rx['value']

                        if vv=='' or vv==None: vv=v

                        mchoices[k][v]=vv

                    wchoices[k].append({'name':vv, 'value':v})

    if debug: h+='\n<p>Debug time (CK find unique vars): '+str(time.time()-dt)+' sec.<p>\n'

    # Prepare query div ***************************************************************
    dt=time.time()
    if cmuoa=='':
        # Start form + URL (even when viewing entry)
        r=ck.access({'action':'start_form',
                     'module_uoa':cfg['module_deps']['wfe'],
                     'url':url1,
                     'name':form_name})
        if r['return']>0: return r
        h+=r['html']

    for kk in selector:
        k=ckey+kk['key']
        n=kk['name']

        nl=kk.get('new_line','')
        if nl=='yes':
            h+='<br>\n<div id="ck_entries_space8"></div>\n'

        v=''
        if i.get(k,'')!='':
            v=i[k]
            kk['value']=v

        # Show hardware
        ii={'action':'create_selector',
            'module_uoa':cfg['module_deps']['wfe'],
            'data':wchoices.get(k,[]),
            'name':k,
            'onchange':conc, 
            'skip_sort':'no',
            'selected_value':v,
            'style':'margin:5px;'}
        r=ck.access(ii)
        if r['return']>0: return r

        h+='<span style="white-space: nowrap"><b>'+n.replace(' ','&nbsp;')+':</b>&nbsp;'+r['html'].strip()+'</span>\n'

    if debug: h+='\n<p>Debug time (prepare selector): '+str(time.time()-dt)+' sec.<p>\n'

    # Check hidden
    if hi_uid!='':
        h+='<input type="hidden" name="highlight_behavior_uid" value="'+hi_uid+'">\n'

    if hi_user!='':
        h+='<input type="hidden" name="highlight_by_user" value="'+hi_user+'">\n'

    h+='<br><br>'

    # Prune list ******************************************************************
    dt=time.time()

    plst=[]

    for q in lst:
        d=q['meta']
        meta=d.get('meta',{})

        # Check selector
        skip=False
        for kk in selector:
            k=kk['key']
            n=kk['name']
            v=kk.get('value','')

            if v!='' and meta.get(k,'')!=v:
                skip=True
                break # FGG added later - should be correct

        if not skip:
            # Process raw results
            arr=d.get('all_raw_results',[])

            for g in arr:
                nn=copy.deepcopy(q)

                ih=g.get('image_height',0)
                iw=g.get('image_width',0)

                it=0
                if ih!=0 and iw!=0:
                    it=ih*iw

                key=str(ih)+' x '+str(iw)

                prd=g.get('prediction','')
                if prd!='':
                    j1=prd.find('\n')
                    if j1>0:
                        j2=prd.find('\n',j1+1)
                        if j2>0:
                            prd=prd[j1:j2]

                # Check timing - currently temporal ugly hack
                t=g.get('time',[])

                tmin=0
                tmax=0

                if len(t)>0:
                    tmin=min(t)/1E3
                    tmax=max(t)/1E3

                nn['extra']={'key':key, 'raw_results':g, 'time_min':tmin, 'time_max':tmax, 'prediction':prd}

                # Check xOpenME timing
                xopenme=g.get('xopenme',{})
                xxopenme={}

                if type(xopenme)==dict:
                   for xk in xopenme:
                       t=xopenme[xk]

                       tmin=0
                       tmax=0

                       if len(t)>0:
                           tmin=min(t)
                           tmax=max(t)

                       xxopenme['xopenme_'+xk+'_min']=tmin
                       xxopenme['xopenme_'+xk+'_max']=tmax

                nn['extra'].update(xxopenme)

                plst.append(nn)

    if debug: h+='\n<p>Debug time (prune entries by user selection): '+str(time.time()-dt)+' sec.<p>\n'

    # Add extra selectors
    h+='<center>\n'

    wchoices9a=[{'name':'throughput', 'value':'throughput'}, {'name':'latency', 'value':'latency'}]
    key9a='sort_by'

    k=ckey+key9a

    v9a=''
    if i.get(k,'')!='':
        v9a=i[k]
    else:
       v9a=wchoices9a[0]['value']

    # Prepare selector
    ii={'action':'create_selector',
        'module_uoa':cfg['module_deps']['wfe'],
        'data':wchoices9a,
        'name':k,
        'onchange':conc, 
        'skip_sort':'yes',
        'selected_value':v9a}
    r=ck.access(ii)
    if r['return']>0: return r

    h+='<b>Sort by:</b> '+r['html'].strip()+'\n'

    wchoices9b=[{'name':'throughput vs top 1 accuracy', 'value':'throughput_top1'},
                {'name':'throughput vs top 5 accuracy', 'value':'throughput_top5'}, 
                {'name':'latency vs top 1 accuracy', 'value':'latency_top1'},
                {'name':'latency vs top 5 accuracy', 'value':'latency_top5'}, 
                {'name':'recognition time vs device cost vs model size', 'value':'time_cost'}]
    key9b='which_graph'

    k=ckey+key9b

    v9b=''
    if i.get(k,'')!='':
        v9b=i[k]
    else:
       v9b=wchoices9b[0]['value']

    # Prepare selector
    ii={'action':'create_selector',
        'module_uoa':cfg['module_deps']['wfe'],
        'data':wchoices9b,
        'name':k,
        'onchange':conc,
        'skip_sort':'yes',
        'selected_value':v9b}
    r=ck.access(ii)
    if r['return']>0: return r

    h+='&nbsp;&nbsp;<b>Plot:</b> '+r['html'].strip()+'\n'

    h+='</center>\n'
    h+='<p>\n'

    # Sort first before prunning
    dt=time.time()

    if v9a=='throughput':
       splst=sorted(plst, key=lambda x: x.get('extra',{}).get('xopenme_execution_time_kernel_2_min',0))
    else:
       splst=sorted(plst, key=lambda x: x.get('extra',{}).get('time_min',0))

    if debug: h+='\n<p>Debug time (sorting table): '+str(time.time()-dt)+' sec.<p>\n'

    # Demo graph
    bgraph={'0':[]} # Just for graph demo
    if hi_uid!='' or hi_user!='': bgraph['1']=[]

    # execution time vs cost
    bgraph2={'0':[]}
    igraph2={'0':[]}

    # execution time vs cost
    bgraph3={'0':[]}
    igraph3={'0':[]}
    f_bgraph3={}
    f_igraph3={}

    # Check if too many
    lplst=len(plst)
    min_view=False

    view_all=i.get('all','')
    if view_all!='': h+='<input type="hidden" name="view_all" value="'+view_all+'">\n'

    # Advertisement
    h+='<center>\n'
#    hhh+=' <a href="http://dividiti.com"><img src="http://cKnowledge.org/_resources/ai-cloud.png" height="240" style="padding:3px;"></a>\n'
#    hhh+=' <iframe width="426" height="240" src="https://www.youtube.com/embed/f4CfMrGPJPY" frameborder="0" style="padding:3px;"></iframe><br>\n'
    h+=' These are raw benchmarking/optimization results shared by volunteers from their Android devices\n'
    h+=' to test our collaborative approach to unify and crowdsource learning, benchmarking, optimization and co-design of efficient SW/HW stack for emerging workloads\n'
    h+=' across diverse devices, models and data sets. See <a href="http://cKnowledge.org/request.html">ACM ReQuEST tournaments</a> for more details!\n'
    h+='</center>\n'
    h+='<br>\n'

    if lplst==0:
        h+='<b>No results found!</b>'
        return {'return':0, 'html':h, 'style':st}
    elif lplst>100 and view_all!='yes':
        h+='<b>Too many <a href="https://arxiv.org/abs/1506.06256">species</a> ('+str(lplst)+') - see 150 top performing species or prune list further!</b><br><br>'

#        min_view=True

        # Removing entries but leave user selection
        splst1=[]
        iq=0
        for q in splst:
            rres=q['extra'].get('raw_results',{})
            buid=rres.get('behavior_uid','')
            user=rres.get('user','')

            if iq<150 or (hi_uid!='' and buid==hi_uid) or (hi_user!='' and hi_user==user):
               splst1.append(q)

            iq+=1

        splst=splst1

#        del splst[150:]

#        return {'return':0, 'html':h, 'style':st}

    # Long term TBD: here we should move table prepation either to numpy or django or both ...
    # This is just for a quick proof-of-concept (FGG)

    # Prepare table
#    bgc='afffaf'
    bgc='dfffdf'
    bg=' style="background-color:#'+bgc+';"'
    bg1=' style="background-color:#bfffbf;"'
    bg2=' style="background-color:#afffaf;"'
    bg9=' style="background-color:#afffaf;color:#7f0000;"'

    h+='<table border="1" cellpadding="7" cellspacing="0">\n'

    ha='align="center" valign="top"'
    hb='align="left" valign="top"'

    h+='  <tr style="background-color:#dddddd">\n'

    h+='   <td '+ha+'><b>#</b></td>\n'
    h+='   <td '+ha+'><b>Platform</b></td>\n'
    h+='   <td '+ha+'><b>Crowd scenario</b></td>\n'
    if not min_view:
       h+='   <td '+ha+'><b>Versions</b></td>\n'
    h+='   <td '+ha+'><b>Model weight size</b></td>\n'
    x=''
    if v9a=='latency':
       x=' style="color:#7f0000"'
    h+='   <td '+ha+x+'><b>Total time (min/max sec.)<p>*&nbsp;LATENCY&nbsp;*</b></td>\n'
    if not min_view:
       h+='   <td '+ha+'><b>Init network time (min/max sec.)</b></td>\n'
       h+='   <td '+ha+'><b>Image preparation (min/max sec.)</b></td>\n'
       x=''
       if v9a=='throughput':
          x=' style="color:#7f0000"'
       h+='   <td '+ha+x+'><b>Classification time (min/max sec.)<p>*&nbsp;THROUGHPUT&nbsp;*</b></td>\n'
       h+='   <td '+ha+'><b>Prediction probability</b></td>\n'
    h+='   <td '+ha+'><b>Power consumption (W)<br>min / max</td>\n'
    h+='   <td '+ha+'><b>Memory usage (MB)</td>\n'
    h+='   <td '+ha+'><b><a href="https://github.com/dividiti/ck-caffe/blob/master/script/explore-accuracy/explore_accuracy.20160808.ipynb">Model accuracy on ImageNet</a></td>\n'
    h+='   <td '+ha+'><b>Model topology and parameters</td>\n'
    h+='   <td '+ha+'><b>HW costs</td>\n'
    h+='   <td '+ha+'><b>All usage costs (preparation, training, inference, errors, etc)</td>\n'
    if not min_view:
       h+='   <td '+ha+'><b>Mispredictions and unexpected behavior</b></td>\n'
       h+='   <td '+ha+'><b>Image features</b></td>\n'
    h+='   <td '+ha+'><b>CPU</b></td>\n'
    if not min_view:
       h+='   <td '+ha+'><b>CPU ABI</b></td>\n'
    h+='   <td '+ha+'><b>GPU</b></td>\n'
    h+='   <td '+ha+'><b>OS</b></td>\n'
    h+='   <td '+ha+'><b>Data UID / Behavior UID</b></td>\n'
    if not min_view:
       h+='   <td '+ha+'><b>User</b></td>\n'
    h+='  <tr>\n'

    # Dictionary to hold target meta
    tm={}

    ix=0

    # Trick - want to put graph first
    hhh=h
    h=''

    dt=time.time()
    for q in splst:
        ix+=1

        duid=q['data_uid']
        path=q['path']

        d=q['meta']

        meta=d.get('meta',{})

        extra=q['extra']
        img=extra.get('img',{})
        rres=extra.get('raw_results',{})

        mp=rres.get('mispredictions',[])

        key=extra.get('key','')

        buid=rres.get('behavior_uid','')
        pred=extra.get('prediction','').replace(' ','&nbsp;')
        user=rres.get('user','')

        tmin=extra.get('time_min',0)
        tmax=extra.get('time_max',0)

        scenario=meta.get('crowd_uid','')

        plat_name=meta.get('plat_name','')
        cpu_name=meta.get('cpu_name','')
        cpu_abi=meta.get('cpu_abi','')
        os_name=meta.get('os_name','')
        gpu_name=meta.get('gpu_name','')
        gpgpu_name=meta.get('gpgpu_name','')

        plat_uid=meta.get('platform_uid','')
        cpu_uid=meta.get('cpu_uid','')
        os_uid=meta.get('os_uid','')
        gpu_uid=meta.get('gpu_uid','')
        gpgpu_uid=meta.get('gpgpu_uid','')

        te=d.get('characteristics',{}).get('run',{})

        bgx=bg
        bgx1=bg1
        bgx2=bg2
        bgx9=bg9

        if (hi_uid!='' and buid==hi_uid) or (hi_user!='' and hi_user==user):
           bgx=' style="background-color:#ffcf7f"'
           bgx1=' style="background-color:#ffbf5f"'
           bgx2=' style="background-color:#ffaf2f"'
           bgx9=' style="background-color:#ffaf2f;color:#7f0000;"'

        # Starting raw
        h+='  <tr'+bgx+'>\n'

        h+='   <td '+ha+'><a name="'+buid+'" id="'+buid+'">'+str(ix)+'</a></td>\n'

        x=plat_name
        if plat_uid!='':
            x='<a href="'+url0+'&wcid='+cfg['module_deps']['platform']+':'+plat_uid+'">'+x+'</a>'
        h+='   <td '+ha+'>'+x+'</td>\n'

        # Output scenario
        xx=mchoices.get(ckey+'crowd_uid',{}).get(scenario,'')

        kscenario=cfg['module_deps']['experiment.scenario.mobile']+':'+scenario
        xd=cache_meta[kscenario]

        xx=xd.get('title','')
        model_weights_size=int(xd.get('model_weights_size',0))+1

        fbsize=xd.get('features',{}).get('fixed_batch_size','')

        h+='   <td '+ha+'><a href="'+url0+'&wcid='+kscenario+'">'+xx+'</a></td>\n'

        # Versions
        if not min_view:
           ver=''
           dver=meta.get('engine_meta',{}).get(cpu_abi,{})
           ver+='main: '+str(dver.get('program_version',''))+'\n'
           dps=dver.get('deps_versions',{})
           for dx in dps:
               ver+=dx+': '+str(dps[dx].get('version',''))+'\n'

           ver=ver.replace("\'","'").replace("'","\\'").replace('\"','"').replace('"',"\\'").replace('\n','\\n')
           if ver!='':
               ver='<input type="button" class="ck_small_button" onClick="alert(\''+ver+'\');" value="View">'
           h+='   <td '+ha+'>'+ver+'</td>\n'

        # Model weight size
        h+='   <td '+ha+'>'+str(model_weights_size)+' MB</td>\n'

        # Check relative time
        xx='<b>'+('%.3f'%tmin)+'</b>&nbsp;/&nbsp;'+('%.3f'%tmax)

        if tmin==0: xx+='<br><b><center>bug?</center></b>\n'
#        TBD: need to check how time is normalized ...
#        elif fbsize!='':
#           xx+='<br><br><i>batch size='+str(fbsize)+'</i>'

        if (hi_uid!='' and buid==hi_uid) or (hi_user!='' and hi_user==user):
            bgraph['0'].append([ix,None])
            bgraph['1'].append([ix,tmin])
        else:
            bgraph['0'].append([ix,tmin])
            if hi_uid!='' or hi_user!='': 
               bgraph['1'].append([ix,None])

        x=bgx9 if v9a=='latency' else bgx1

        h+='   <td '+ha+' '+x+'>'+xx+'</a></td>\n'

        # Finer grain timing
        ttmin=tmin
        ttdelta=tmax-tmin

        if not min_view:
           for ixo in range(0,3):
              tmin=extra.get('xopenme_execution_time_kernel_'+str(ixo)+'_min',0)
              tmax=extra.get('xopenme_execution_time_kernel_'+str(ixo)+'_max',0)

              if tmin<0: tmin=0 # detected bug

              xx='<b>'+('%.3f'%tmin)+'</b>&nbsp;/&nbsp;'+('%.3f'%tmax)
              if tmin==0 and ixo!=1: 
                 xx+='<br><b><center>possible bug detected - check further</center></b>\n'

              x=bgx9 if (v9a=='throughput' and ixo==2) else bgx1

              if v9b.startswith('throughput') and ixo==2:
                 ttmin=tmin
                 ttdelta=tmax-tmin

              h+='   <td '+ha+' '+x+'>'+xx+'</a></td>\n'

           # Accuracy
           x=pred
           j=x.find('-')
           if j>0:
               x=x[:j-1].strip()
           h+='   <td '+ha+' '+bgx2+'>'+x+'</a></td>\n'

        # Get info about platform
        hd={}
        if plat_uid!='':
           rh=ck.access({'action':'load',
                        'module_uoa':cfg['module_deps']['platform'],
                        'data_uoa':plat_uid})
           if rh['return']==0:
              hd=rh['dict']

        # Energy TBD
        x='-'
        if len(hd)>0:
           power=hd.get('features',{}).get('power_consumption',{})
           if len(power)>0:
              pmin=power.get('min','')
              pmax=power.get('max','')

              x=str(pmin)+' / '+str(pmax)
        h+='   <td '+ha+'>'+x+'</a></td>\n'

        # Memory usage TBD
        h+='   <td '+ha+'>-</a></td>\n'

        # Accuracy (take from model info)
        acc=xd.get('features',{}).get('accuracy','')
        acc5=xd.get('features',{}).get('accuracy_top5','')

        x=str(acc)
        if acc5!='':
           x+=' / '+str(acc5)

        if x=='': x='-'

        h+='   <td '+ha+'>'+x+'</a></td>\n'

        # Will be used to optimize model topology and parameters
        x='default'

        xfiles=xd.get('files',[])
        for xf in xfiles:
            if xf.get('filename','')=='deploy.prototxt':
               xx1=xf.get('from_data_uoa','')
               if xx1=='':
                  xx1=scenario
               xx2=xf.get('path','')

               x='<a href="'+url0+'&action=pull&common_action=yes&cid='+cfg['module_deps']['experiment.scenario.mobile']+':'+xx1+'&filename='+xx2+'/deploy.prototxt'+'">deploy.prototxt</a>\n'

               break

        h+='   <td '+ha+'>'+x+'</a></td>\n'

        # Cost (take from platform meta)
        hc='-'
        last_cost=0
        if len(hd)>0:
           costs=hd.get('features',{}).get('cost',[])
           hc=''
           if len(costs)>0:
              last_cost=int(costs[0].get('price','0'))
              for c in costs:
                  if hc!='': hc+='<br>\n'
                  hc+='<b>'+str(c.get('price',''))+' '+c.get('currency','')+ '</b> - '+c.get('desc','')+' ('+c.get('date','')+')'

        h+='   <td '+ha+'>'+hc+'</a></td>\n'

        # TBD: all other costs
        h+='   <td '+ha+'></a></td>\n'

        # Mispredictions and unexpected behavior
        if not min_view:
           x=''
           for q in mp:
               ca=q.get('correct_answer','')
               mi=q.get('mispredicted_image','')
               mr=q.get('misprediction_results','')

               if mr!='':
                   j1=mr.find('\n')
                   if j1>0:
                       j2=mr.find('\n',j1+1)
                       if j2>0:
                           mr=mr[j1:j2]

               xx=ca
               if mi!='':
                   y=work['self_module_uid']
                   if cmuoa!='': y=cmuoa
                   url=url0+'action=pull&common_action=yes&cid='+y+':'+duid+'&filename='+mi

                   if ca=='': ca='<i>unknown</i>'
                   xx='<a href="'+url+'">'+ca+'</a>'

               if x!='':
                   x+='<hr>\n'

               x+='<strike>'+mr+'</strike><br>'+xx+'<br>\n'

           if tmin==0: x+='<br><b><center>Bug detected</center></b>\n'

           h+='   <td '+ha+'>'+x+'</td>\n'

           # All images
           h+='   <td '+ha+'>'+key.replace(' ','&nbsp;')+'</a></td>\n'

        # Extra info about platform
        x=cpu_name
        if cpu_uid!='':
            x='<a href="'+url0+'&wcid='+cfg['module_deps']['platform.cpu']+':'+cpu_uid+'">'+x+'</a>'
        h+='   <td '+ha+'>'+x+'</td>\n'

        # CPU ABI
        if not min_view:
           h+='   <td '+ha+'>'+cpu_abi+'</td>\n'

        x=gpu_name
        if gpu_uid!='':
            x='<a href="'+url0+'&wcid='+cfg['module_deps']['platform.gpu']+':'+gpu_uid+'">'+x+'</a>'
        h+='   <td '+ha+'>'+x+'</td>\n'

        x=os_name
        if os_uid!='':
            x='<a href="'+url0+'&wcid='+cfg['module_deps']['platform']+':'+os_uid+'">'+x+'</a>'
        h+='   <td '+ha+'>'+x+'</td>\n'

        # Data
        x=work['self_module_uid']
        if cmuoa!='': x=cmuoa
        raw_data_url=url0+'wcid='+x+':'+duid
        h+='   <td '+ha+'><a href="'+raw_data_url+'">'+duid+' '+buid+'</a></td>\n'

        # User
        if not min_view:
           h+='   <td '+ha+'><a href="'+url0+'&action=index&module_uoa=wfe&native_action=show&native_module_uoa=experiment.user">'+user+'</a></td>\n'

           h+='  <tr>\n'

        tdelta=tmax-tmin # show best specie!
#        if tdelta>0.1: tdelta=0.1 # to avoid messy graph - just show that there is an issue with variation ...

        if tmin>0: # to skip bugs
           # check accuracy
           xcol='#0000bf'
           size5=2
           if acc5!='' and acc5>0:
              top5=(float(acc5)-0.8)*1000+50
              size5=(float(acc5)-0.75)*40+2
              col=hex(int(top5))[2:]
              xcol='#'+col+col+'ff'

           # mem usage
##           xx=(int(model_weights_size/20))*5.35+127
#           x=hex(int((xx)))[2:]
#           if len(x)<2: x='0'+x
#           mcol='#0000'+x

           divide=80
           if i.get(ckey+'engine')=='TFLite CPU':
              divide=3
           sizem=int(1+(model_weights_size/divide))+3

           bgraph2['0'].append([model_weights_size,tmin,tmin+tdelta])
           igraph2['0'].append({'size':3,'color':xcol})

           features={}

           X=0
           Y=0
           YD=0
           YS=0
           to_add=False

           if v9b=='time_cost':
              if last_cost>0:
                 to_add=True

                 if debug:
                    sizem=int(1+(model_weights_size/80))*6

                 X=last_cost
                 Y=tmin
                 YD=tdelta
                 YS=sizem
           else:
              if acc>0 and ttmin>0:
                 to_add=True

                 if debug:
                    sizem=int(1+(model_weights_size/80))*6

                 X=ttmin
                 if v9b.endswith('top1'):
                    Y=acc
                 else:
                    Y=acc5
                 YD=ttdelta
                 YS=sizem

                 if Y=='': 
                    Y=0
                    YD=0

           if to_add:
              # FGG: must improve to make a constant threshold afterwards (after knowing Ymin and Ymax)
              if (type(Y)==float or type(Y)==int) and Y>0 and YD>0 and YD>float(Y)*0.05: YD=float(Y)*0.05

              bgraph3['0'].append([X,Y,Y+YD])
#              igraph3['0'].append({'size':YS, 'color':xcol, 'features':meta, 'url':'', 'url_ext':raw_data_url})
              igraph3['0'].append({'size':YS, 'color':xcol, 'features':meta, 'anchor':buid})

              # For frontier
#              rx=ck.gen_uid({})
#              if rx['return']>0: return rx
#              puid=rx['data_uid']

#              f_bgraph3[puid]={"X":X, "Y":Y}
#              f_igraph3[puid]={'size':YS, 'tdelta':YD, 'features':meta, 'url':'', 'url_ext':raw_data_url}


    h+='</table>\n'
    h+='</center>\n'

    if debug: h+='\n<p>Debug time (preparing html of a table): '+str(time.time()-dt)+' sec.<p>\n'

    if cmuoa=='':
        h+='</form>\n'

    # 1st graph
    if len(bgraph['0'])>0:
       dt=time.time()
       ii={'action':'plot',
           'module_uoa':cfg['module_deps']['graph'],

           "table":bgraph,

           "ymin":0,

           "ignore_point_if_none":"yes",

           "plot_type":"d3_2d_bars",

           "display_y_error_bar":"no",

           "title":"Powered by Collective Knowledge",

           "x_ticks_period":50,

           "axis_x_desc":"Experiment",
           "axis_y_desc":"DNN image classification time (s)",

           "plot_grid":"yes",

           "d3_div":"ck_interactive",

           "image_width":"900",
           "image_height":"400",

           "wfe_url":url0}

#       r=ck.access(ii)
#       if r['return']==0:
#          x=r.get('html','')
#          if x!='':
#             st+=r.get('style','')
#
#             h+='<br>\n'
#             h+='<center>\n'
#             h+='<div id="ck_box_with_shadow" style="width:920px;">\n'
#             h+=' <div id="ck_interactive" style="text-align:center">\n'
#             h+=x+'\n'
#             h+=' </div>\n'
#             h+='</div>\n'
#             h+='</center>\n'

    # 2nd graph
    if len(bgraph2['0'])>0:
       ii={'action':'plot',
           'module_uoa':cfg['module_deps']['graph'],

           "table":bgraph2,
           "table_info":igraph2,

           "xmin":-10,
           "ymin":0,

           "ignore_point_if_none":"yes",

           "plot_type":"d3_2d_scatter",

           "display_y_error_bar2":"no",

           "title":"Powered by Collective Knowledge",

           "x_ticks_period":50,

#           "axis_x_desc":"Weigths size (MB)",
#           "axis_y_desc":"DNN image classification time (s)",

           "plot_grid":"yes",

           "d3_div":"ck_interactive2",

           "image_width":"900",
           "image_height":"400",

           "wfe_url":url0}

#       r=ck.access(ii)
#       if r['return']==0:
#          x=r.get('html','')
#          if x!='':
#             st+=r.get('style','')
#
#             h+='<br>\n'
#             h+='<center>\n'
#             h+='<div id="ck_box_with_shadow" style="width:920px;">\n'
#             h+='\n'
#             h+=' <div id="ck_interactive2" style="text-align:center">\n'
#             h+=x+'\n'
#             h+=' </div>\n'
#             h+='</div>\n'
#             h+='</center>\n'

    # 3nd graph
    graph_html=''
    graph_style=''

    if len(bgraph3['0'])>0:
       # Fontier
       r=ck.access({'action':'filter',
                    'module_uoa':cfg['module_deps']['math.frontier'],
                    'points':f_bgraph3})
       if r['return']>0: return r

       points=r['points']

       bgraph3['1']=[]
       igraph3['1']=[]

       if v9b=='time_cost':
          xdesc='Device price, $'
          ydesc='image classification time, sec.'
       elif v9b=='throughput_top1' or v9b=='':
          xdesc='throughput (image classification), sec.'
          ydesc='Top1 accuracy'
       elif v9b=='throughput_top5':
          xdesc='throughput (image classification), sec.'
          ydesc='Top5 accuracy'
       elif v9b=='latency_top1':
          xdesc='latency (image classification), sec.'
          ydesc='Top1 accuracy'
       elif v9b=='latency_top5':
          xdesc='latency (image classification), sec.'
          ydesc='Top5 accuracy'

       for q in points:
           q1=f_bgraph3[q]
           q2=f_igraph3[q]

           bgraph3['1'].append([q1['X'],q1['Y'],q1['Y']+q2['tdelta']])
           del(q2['tdelta'])
           igraph3['1'].append(q2)

       ii={'action':'plot',
           'module_uoa':cfg['module_deps']['graph'],

           "table":bgraph3,
           "table_info":igraph3,

           "ignore_point_if_none":"yes",

           "plot_type":"d3_2d_scatter",
           "display_y_error_bar2":"yes",

           "title":"Powered by Collective Knowledge",

           "x_ticks_period":50,

           "axis_x_desc":xdesc,
           "axis_y_desc":ydesc,

            "xmin":-0.01,
            "ymin":0,
#           "ymax":25,

           "point_style":{"1":{"color":"#dc3912", "connect_lines":"no"}},

           "plot_grid":"no",

           "d3_div":"ck_interactive3",

           "image_width":"900",
           "image_height":"400",

           "wfe_url":url0}

       # Temporal hack
       if debug: 
          ii['out_to_file']='/tmp/dnn-co-design.pdf'
          ii['customize_dots']='yes'
          ii['plot_type']='mpl_2d_scatter'

       r=ck.access(ii)
       if r['return']==0:
          graph_html=r.get('html','')
          if graph_html!='':
             graph_style=r.get('style','')
             st+=graph_style

             hhh+='<center>\n'
             hhh+='<div id="ck_box_with_shadow" style="width:920px;">\n'
             hhh+=' <div id="ck_interactive3" style="text-align:center">\n'
#             hhh+='Device cost (X axis, &euro;) vs image classification time (Y axis, s.) vs model size (dot size) vs model TOP5 ImageNet accuracy (darker colors for lower accuracy).\n'
             hhh+='Dot size ~ model size ; dot color ~ model TOP5 ImageNet accuracy (darker colors for lower accuracy).\n'
#             hhh+='<b><span style="color:#9f0000">Red dots - surviving species for a given realistic AI scenario</span></b>.<br>\n'
             hhh+=graph_html+'\n'
             hhh+=' </div>\n'
             hhh+='</div>\n'
             hhh+='</center>\n'
             hhh+='<br>\n'

       # Extra vision slide
       hhh+='<p>\n<center><a href="https://www.slideshare.net/GrigoriFursin/adapting-to-a-cambrian-aiswhw-explosion-with-open-codesign-competitions-and-collective-knowledge"><img src="http://cknowledge.org/_resources/b4b07ad3a7839327-cropped.png" width="600"></a></center><p>\n'

       if debug: hhh+='\n<p>Debug time (preparing graph): '+str(time.time()-dt)+' sec.<p>\n'

    # Adding table
    hhh+=h

    rr={'return':0, 'style':st}

    if minimal:
       rr['graph_html']=graph_html
       rr['style']=graph_style
       rr['table']=bgraph3
       rr['table_info']=igraph3
    else:
       rr['html']=hhh

    return rr

##############################################################################
# process raw results from mobile devices

def process(i):
    """
    Input:  {
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    import copy

#    Debug
#    ck.save_json_to_file({'json_file':'/tmp/xyz1.json','dict':i})

    crowd_uid=i.get('crowd_uid','')
    email=i.get('email','')
    raw_results=i.get('raw_results',{})

    cfb=i.get('cpu_freqs_before',{})
    cfa=i.get('cpu_freqs_after',{})

    features=i.get('platform_features',{})

    fplat=features.get('platform',{})
    fos=features.get('os',{})
    fcpu=features.get('cpu',{})
    fgpu=features.get('gpu',{})

    plat_name=fplat.get('name','')
    plat_uid=features.get('platform_uid','')
    os_name=fos.get('name','')
    os_uid=features.get('os_uid','')
    gpu_uid=features.get('gpu_uid','')
    cpu_name=fcpu.get('name','')
    cpu_abi=fcpu.get('cpu_abi','')
    if cpu_name=='' and cpu_abi!='': 
        cpu_name='unknown-'+fcpu.get('cpu_abi','')
    cpu_uid=features.get('cpu_uid','')
    gpu_name=fgpu.get('name','')
    gpgpu_name=''
    gpgpu_uid=''
    sn=fos.get('serial_number','')

    # Prepare high-level experiment meta
    meta={'cpu_name':cpu_name,
          'cpu_abi':cpu_abi,
          'os_name':os_name,
          'plat_name':plat_name,
          'gpu_name':gpu_name,
          'gpgpu_name':gpgpu_name,
          'crowd_uid':crowd_uid}

    # Load scenario and update meta
    r=ck.access({'action':'load',
                 'module_uoa':cfg['module_deps']['experiment.scenario.mobile'],
                 'data_uoa':crowd_uid})
    if r['return']>0: return r
    xd=r['dict']

    meta['engine']=xd.get('engine','')
    meta['model']=xd.get('model','')

    mmeta=copy.deepcopy(meta)

    # Extra meta which is not used to search similar case ...
    mmeta['platform_uid']=plat_uid
    mmeta['os_uid']=os_uid
    mmeta['cpu_uid']=cpu_uid
    mmeta['gpu_uid']=gpu_uid
    mmeta['gpgpu_uid']=gpgpu_uid

    # Generate behavior UID
    rx=ck.gen_uid({})
    if rx['return']>0: return rx
    buid=rx['data_uid']

    raw_results['user']=email
    raw_results['behavior_uid']=buid

    # Check if already exists
    duid=''
    ddd={}

    ii={'action':'search',
        'module_uoa':work['self_module_uid'],
        'repo_uoa':ck.cfg.get('record_local_repo_uoa',''),
        'search_dict':{'meta':meta},
        'add_meta':'yes'}
    rx=ck.access(ii)
    if rx['return']>0: return rx

    lst=rx['lst']

    if len(lst)==1:
        duid=lst[0]['data_uid']
        ddd=lst[0]['meta']
    else:
        rx=ck.gen_uid({})
        if rx['return']>0: return rx
        duid=rx['data_uid']

    # We keep time1,2,3 just for compatibility with the first beta version
    t=raw_results.get('time',[])
    tx=raw_results.get('time1',None)
    if tx!=None: t.append(tx)
    tx=raw_results.get('time2',None)
    if tx!=None: t.append(tx)
    tx=raw_results.get('time3',None)
    if tx!=None: t.append(tx)
    raw_results['time']=t

    # Check XOpenME
    xopenme={}

    yopenme=raw_results.get('xopenme',[])
    for xx in yopenme:
        for k in xx:
            v=xx[k]
            if k not in xopenme:
               xopenme[k]=[]
            if v!=None:
               xopenme[k].append(v)

    raw_results['xopenme']=xopenme

    # Record freq (not checking if changed at this stage)
    raw_results['cpu_freqs_before']=[cfb]
    raw_results['cpu_freqs_after']=[cfa]

    # Process freq before and freq after (for now no any intelligence)
    fb=raw_results

    # Process results
    results=ddd.get('all_raw_results',[])

    # Check if already exists with this image topology
    found=False
    for q in results:
        if (q.get('image_height',None)==raw_results.get('image_height',None) and \
           q.get('image_width',None)==raw_results.get('image_width',None)) or \
           (q.get('image_height',None)==raw_results.get('image_width',None) and \
           q.get('image_width',None)==raw_results.get('image_height',None)):
            t=q.get('time',[])

            for tx in raw_results.get('time',[]):
                t.append(tx)
            q['time']=t

            tkk1=raw_results.get('xopenme',{})
            tkk2=q.get('xopenme',{})

            for tk in tkk1:
                if tk not in tkk2: tkk2[tk]=[]
                for tx in tkk1[tk]:
                    if tx!=None: 
                       tkk2[tk].append(tx)

            q['xopenme']=tkk2

            fb=q.get('cpu_freqs_before',[])
            fb.append(cfb)
            q['cpu_freqs_before']=fb

            fa=q.get('cpu_freqs_after',[])
            fa.append(cfa)
            q['cpu_freqs_before']=fa

            buid=q.get('behavior_uid','')

            found=True

            break

    if not found:
        results.append(raw_results)

    ddd['all_raw_results']=results

    xmeta=ddd.get('meta',{})
    xmeta.update(mmeta)
    ddd['meta']=xmeta

    # Update meta
    rx=ck.access({'action':'update',
                  'module_uoa':work['self_module_uid'],
                  'data_uoa':duid,
                  'repo_uoa':ck.cfg.get('record_local_repo_uoa',''),
                  'dict':ddd,
                  'substitute':'yes',
                  'sort_keys':'yes'})
    if rx['return']>0: return rx

    # Prepare url with results
    rx=ck.access({'action':'form_url_prefix',
                  'module_uoa':'wfe'})
    if rx['return']>0: return rx

    url=rx['url']+'&action=index&module_uoa=wfe&native_action=show&native_module_uoa=program.optimization&scenario='+work['self_module_uid']+'&highlight_behavior_uid='+buid

    return {'return':0, 'status':'Results successfully added to Collective Knowledge (UID='+duid+')!', 'data_uid':duid, 'behavior_uid':buid, 'result_url':url}

##############################################################################
# record unexpected behavior

def process_unexpected_behavior(i):
    """
    Input:  {
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    import os

    duid=i.get('data_uid','')
    buid=i.get('behavior_uid','')
    cuid=i.get('crowd_uid','')
    rres=i.get('raw_results','')
    ca=i.get('correct_answer','')
    file_base64=i.get('file_base64','')

    # Find data
    r=ck.access({'action':'load',
                 'module_uoa':work['self_module_uid'],
                 'data_uoa':duid})
    if r['return']>0: return r
    d=r['dict']
    p=r['path']

    # Find behavior
    found=False
    arr=d.get('all_raw_results',[])
    for q in arr:
        if q.get('behavior_uid','')==buid:
            found=True
            break

    if not found:
        return {'return':1, 'error':'can\'t find behavior '+buid+' in entry '+duid}

    # Generate UID for the file with unexpected behavior
    rx=ck.gen_uid({})
    if rx['return']>0: return rx

    ff='misprediction-image-'+rx['data_uid']+'.jpg'

    pf=os.path.join(p,ff)

    mp=q.get('mispredictions',[])

    qq={}
    qq['misprediction_results']=rres
    qq['mispredicted_image']=ff
    qq['correct_answer']=ca

    mp.append(qq)

    q['mispredictions']=mp

    # Record file
    rx=ck.convert_upload_string_to_file({'file_content_base64':file_base64,
                                         'filename':pf})
    if rx['return']>0: return rx

    # Update entry (should add lock in the future for parallel processing)
    r=ck.access({'action':'update',
                 'module_uoa':work['self_module_uid'],
                 'data_uoa':duid,
                 'dict':d,
                 'sort_keys':'yes',
                 'substitute':'yes',
                 'ignore_update':'yes'})
    if r['return']>0: return r

    return {'return':0}

##############################################################################
# generate scenario to crowdsource benchmarking/optimization of DNN engines/models on Android devices

def generate(i):
    """
    Input:  {
              (prune_target_os) - prune generated scenarios by this target OS (ABI)
              (prune_engine)    - prune generated scenarios by this engine
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    import os
    import copy
    import shutil
    import itertools

    o=i.get('out','')

    p_tos=i.get('prune_target_os','')
    p_engine=i.get('prune_engine','')

    # Get platform params
    ck.out('Detecting host platform info ...')
    ck.out('')

    i['action']='detect'
    i['module_uoa']=cfg['module_deps']['platform']
    i['out']=''
    rpp=ck.access(i)
    if rpp['return']>0: return rpp

    hos=rpp['host_os_uoa']
    hosd=rpp['host_os_dict']

    # Search all scenarios
    r=ck.access({'action':'search',
                 'module_uoa':cfg['module_deps']['experiment.scenario.mobile'],
                 'add_meta':'yes'})
    if r['return']>0: return r

    lst=r['lst']
    if len(lst)==0:
       return {'return':1, 'error':'no pre-recorded crowd-scenarios found ...'}

    uids={}
    found=False

    engine_meta={}

    engine_state='' # If engine changes, clean up found uids ...

    # Go through required engines (CPU,OpenCL) and ABIs, and compile classification with related libs ...
    for prog,b in list(itertools.product(cfg['prog_uoa'],abis)):

        abi=b['abi']
        os_uoa=b['os_uoa']
        if i.get('target_os','')!='': os_uoa=i['target_os']

        engine=prog['engine']
        prog_uoa=prog['program_uoa']
        engine_key=prog['engine-deps-key']
        engine_lib=prog.get('engine-lib','')

        # Check if should prune
        if p_tos!='' and p_tos!=os_uoa:
           continue

        if p_engine!='' and p_engine!=engine:
           continue

        if engine!=engine_state:
           uids={}
           found=False
           engine_state=engine

        ck.out(line)
        ck.out('Preparing "'+engine+'" engine with "'+abi+'" ABI for crowd-benchmarking and crowd-tuning ...')

        ck.out('')
        r=ck.inp({'text':'Press Enter to generate this scenario or N to skip this scenario: '})
        s=r['string']

        if s!='':
           continue

        # Compile classification (should compile all deps)
        ck.out('')
        ck.out('Compiling classification (and libs) ...')

        ii={'action':'compile',
            'module_uoa':cfg['module_deps']['program'],
            'data_uoa':prog_uoa,
            'host_os':hos,
            'target_os':os_uoa,
            'speed':'yes',
            'out':o}
        r=ck.access(ii)
        if r['return']>0: return r

        # Get various info
        cc=r.get('characteristics',{})

        cs=cc.get('compilation_success','')
        if cs!='yes':
           return {'return':1, 'error':'compilation failed'}

        md5=cc.get('md5_sum','')
        bs=cc.get('binary_size',0)

        misc=r.get('misc',{})

        pf=misc.get('path','')
        td=misc.get('tmp_dir','')
        of=misc.get('target_exe','')
        pv=misc.get('program_version','')

        pp=os.path.join(pf,td,of)
        if not os.path.isfile(pp):
           return {'return':1, 'error':'target binary not found ('+pp+')'}

        # Get info about deps
        deps=r.get('deps',{})

        # Get info about lib if needed
        cdeps=deps.get(engine_key,{})

        cfp=''
        lmd5=''
        ls=0

        if engine_lib!='':
           # Force .so (sometimes points to .a)
           cfp=cdeps.get('cus',{}).get('full_path','')
           cfp=os.path.join(os.path.dirname(cfp),engine_lib)

           if os.path.isfile(cfp):
              r=ck.run_and_get_stdout({'cmd':hosd['md5sum']+' '+cfp, 
                                    'shell':'no'})
              if r['return']>0: return r
              sto=r['stdout'].split(' ')

              if len(sto)==0:
                 return {'return':1, 'error':'can\'t get MD5 of '+engine_lib}

              lmd5=sto[0] # MD5 of caffe lib

              ls=os.path.getsize(cfp)

        # xopenme
        odeps=deps.get('xopenme',{})

        ofp=odeps.get('cus',{}).get('full_path','')

        # Force .so (sometimes points to .a)
        ofp=os.path.join(os.path.dirname(ofp),libxopenme)

        if not os.path.isfile(ofp):
           return {'return':1, 'error':'xopenme plugin not found ('+ofp+')'}

        r=ck.run_and_get_stdout({'cmd':hosd['md5sum']+' '+ofp, 
                                 'shell':'no'})
        if r['return']>0: return r
        sto=r['stdout'].split(' ')

        if len(sto)==0:
           return {'return':1, 'error':'can\'t get MD5 of librtlxopenme.so'}

        omd5=sto[0] # MD5 of caffe lib

        ops=os.path.getsize(ofp)

        # Get versions of all deps
        dv={}
        for x in deps:
            dv[x]={}
            cx=deps[x].get('cus',{})

            dv[x]['version']=cx.get('version','')
            dv[x]['revision']=cx.get('git_info',{}).get('revision','')
            dv[x]['iso_datetime_cut_revision']=cx.get('git_info',{}).get('iso_datetime_cut_revision','')
            dv[x]['tags']=deps[x].get('tags',[])

        # Prepare info pack about this experiment
        meta={'classification_version':pv,
              'deps':dv,
              'host_os':hos,
              'target_os':os_uoa}

        # Search caffe original code with a given ABI
        ck.out('')
        ck.out('Checking / updating scenario files ...')

        changed_files=[] # Which libs were substituted

        for q in lst:
           duid=q['data_uid']
           duoa=q['data_uoa']
           ruoa=q['repo_uid']

           d=q['meta']

           if d.get('outdated','')=='yes': # skip archived entries
              continue

           if d.get('engine','')!=engine:
              continue

           pe=q['path'] # Path to entry

           files=d.get('files',[])

           # Process main binary files (in code directory)
           for ff in files:
               fduoa=ff.get('from_data_uoa','')
               if fduoa=='':
                  sabi=ff.get('supported_abi',[])
                  if abi in sabi:
                     fn=ff.get('filename','')
                     p=ff.get('path','')

                     xfs=ff.get('file_size',0)
                     xmd5=ff.get('md5','')

                     pep=os.path.join(pe,p,fn)

                     if (fn==engine_lib and (not os.path.isfile(pep) or (xmd5!=lmd5 and xfs!=ls))) or \
                        (fn=='classification' and (not os.path.isfile(pep) or (xmd5!=md5 and xfs!=bs))) or \
                        (fn==libxopenme and (not os.path.isfile(pep) or (xmd5!=omd5 and xfs!=ops))):

                        if not found:
                           # If first time, tell that old scenario will be removed!
                           ck.out('')
                           ck.out('WARNING: we found OUTDATED binaries for this scenario')
                           ck.out('         and plan to remove them - make sure that you archived them!')

                           ck.out('')
                           r=ck.inp({'text':'Would you like to proceed and remove outdated binaries (Y/n): '})
                           if r['return']>0: return r

                           s=r['string'].strip().lower()

                           if s!='':
                              return {'return':0}

                           found=True

                        ck.out('')
                        ck.out('Updating '+fn+' in '+duoa+' ('+p+')')

                        oduid=p.split('/')[1]

                        nduid=uids.get(oduid,'')

                        # Generate new UID
                        if nduid=='':
                           r=ck.gen_uid({})
                           if r['return']>0: return r
                           nduid=r['data_uid']

                           uids[oduid]=nduid

                           ck.out('  * New code UID for '+oduid+' : '+nduid)
                        else:
                           ck.out('  * Reusing UID for '+oduid+' : '+nduid)

                        np=os.path.join(pe,'code',nduid,abi)
                        ck.out('  * New path: '+np)

                        # Create new dir and copy file
                        if not os.path.isdir(np):
                           os.makedirs(np)

                        fnp=os.path.join(np,fn)

                        if fn==engine_lib:
                           shutil.copy(cfp, fnp)
                        elif fn=='classification':
                           shutil.copy(pp, fnp)
                        else:
                           shutil.copy(ofp, fnp)

                        # Remove old path
                        px=os.path.join(pe,p)
                        if os.path.isdir(px):
                           ck.out('  * Removing dir: '+px)
                           shutil.rmtree(px)

                        # If whole directory is empty, remove it too
                        px=os.path.join(pe,'code',oduid)
                        if os.path.isdir(px) and len(os.listdir(px))==0:
                           ck.out('  * Removing dir: '+px)
                           shutil.rmtree(px)

                        # Changing meta
                        zp=np
                        if not zp.startswith('code') and not zp.startswith('data'):
                           j1=zp.find('code/')
                           if j1<0:
                              j1=zp.find('data/')
                           if j1>0:
                              zp=zp[j1:]

                        changed_before={'filename':fn,
                                        'from_data_uoa':duid,
                                        'path':p,
                                        'file_size':xfs,
                                        'md5':xmd5}

                        changed_after={'filename':fn,
                                       'from_data_uoa':duid,
                                       'path':zp}

                        if fn==engine_lib:
                           changed_after['file_size']=ls
                           changed_after['md5']=lmd5
                        elif fn=='classification':
                           changed_after['file_size']=bs
                           changed_after['md5']=md5
                        else:
                           changed_after['file_size']=ops
                           changed_after['md5']=omd5

                        ff.update(changed_after)
                        del(ff['from_data_uoa'])

                        # Updating global meta of engine
                        engine_meta[abi]={'program_version':pv, 'deps_versions':dv}

                        d['engine_meta']=engine_meta

                        # Add to changed
                        changed_files.append({'before':changed_before,
                                              'after':changed_after})

                     else:
                        ck.out('')
                        ck.out('*** Scenario binary '+fn+' is up to date! ***')

        # If changed original files, change them in all other meta
        if len(changed_files)>0:

           ck.out('')
           ck.out('Updating all scenarios with new files ...')
           ck.out('')

           for q in lst:
               duid=q['data_uid']
               duoa=q['data_uoa']
               muid=q['module_uid']
               ruoa=q['repo_uid']

               d=q['meta']

               if d.get('outdated','')=='yes': # skip archived entries
                  continue

               if d.get('engine','')!=engine:
                  continue

               ck.out('  * '+duoa)

               d['engine_meta']=engine_meta

               pp=q['path'] # Path to entry

               files=d.get('files',[])

               for ff in files:
                   fduoa=ff.get('from_data_uoa','')

                   if fduoa!='':
                      # Check if needs to be changed ...
                      for c in changed_files:
                          before=c['before']
                          after=c['after']

                          rx=ck.compare_dicts({'dict1':ff, 'dict2':before})
                          if rx['return']>0: return rx
                          if rx['equal']=='yes':
                             ff.update(after)

               # Update entry now (aggregating all above changes)
               r=ck.access({'action':'update',
                            'module_uoa':muid,
                            'data_uoa':duid,
                            'repo_uoa':ruoa,
                            'dict':d,
                            'substitute':'yes',
                            'ignore_update':'yes',
                            'sort_keys':'yes'})
               if r['return']>0: return r

    return {'return':0}



#
# Collective Knowledge
#
# See CK LICENSE.txt for licensing details
# See CK COPYRIGHT.txt for copyright details
#

import unittest

ck=None           # Will be updated by CK (initialized CK kernel)
work=None         # Will be updated by CK (initialized CK kernel)
test_util=None    # Will be updated by CK (initialized CK test utils)

# Contains new kernel tests. Add new tests here!
class TestOs(unittest.TestCase):

    def test_lib_path_export_script(self):
        h = { 
            "env_ld_library_path": "DYLD_LIBRARY_PATH",
            "env_library_path": "LIBRARY_PATH",
        }
        dyn_path = '/dyn/path'
        stat_path = '/stat/path'

        r = ck.access({
                'action': 'lib_path_export_script',
                'module_uoa': 'os',
                'host_os_dict': h,
                'dynamic_lib_path': dyn_path,
                'static_lib_path': stat_path
            })
        self.assertEqual(0, r['return'])
        script = '\nexport {dyn_name}="{dyn_val}":${dyn_name}\nexport {stat_name}="{stat_val}":${stat_name}\n\n'.format(
            dyn_name = h['env_ld_library_path'], dyn_val = dyn_path,
            stat_name = h['env_library_path'], stat_val = stat_path)
        self.assertEqual(script, r['script'])

        r = ck.access({
                'action': 'lib_path_export_script',
                'module_uoa': 'os',
                'host_os_dict': h,
                'lib_path': dyn_path
            })
        self.assertEqual(0, r['return'])
        script = '\nexport {dyn_name}="{dyn_val}":${dyn_name}\nexport {stat_name}="{stat_val}":${stat_name}\n\n'.format(
            dyn_name = h['env_ld_library_path'], dyn_val = dyn_path,
            stat_name = h['env_library_path'], stat_val = dyn_path)
        self.assertEqual(script, r['script'])

        h = { 
            "env_library_path": "LIBRARY_PATH",
        }
        r = ck.access({
                'action': 'lib_path_export_script',
                'module_uoa': 'os',
                'host_os_dict': h,
                'dynamic_lib_path': dyn_path
            })
        self.assertEqual(0, r['return'])
        script = '\nexport {dyn_name}="{dyn_val}":${dyn_name}\n\n'.format(
            dyn_name = 'LD_LIBRARY_PATH', dyn_val = dyn_path)
        self.assertEqual(script, r['script'])

        h = { 
            "env_library_path": "LIBRARY_PATH",
            "ck_name": "win"
        }
        r = ck.access({
                'action': 'lib_path_export_script',
                'module_uoa': 'os',
                'host_os_dict': h,
                'dynamic_lib_path': dyn_path
            })
        self.assertEqual(0, r['return'])
        self.assertEqual('', r['script'])

        r = ck.access({
                'action': 'lib_path_export_script',
                'module_uoa': 'os',
                'host_os_dict': {},
                'lib_path': ''
            })
        self.assertEqual(0, r['return'])
        self.assertEqual('', r['script'])

        r = ck.access({
                'action': 'lib_path_export_script',
                'module_uoa': 'os',
                'host_os_dict': {},
                'lib_path': ['a', 'b']
            })
        self.assertEqual(0, r['return'])
        self.assertEqual('\nexport LD_LIBRARY_PATH="a":"b":$LD_LIBRARY_PATH\nexport LIBRARY_PATH="a":"b":$LIBRARY_PATH\n\n', r['script'])



#
# Collective Knowledge (preparing experiment reports)
#
# See CK LICENSE.txt for licensing details
# See CK COPYRIGHT.txt for copyright details
#
# Developer: Grigori Fursin, Grigori.Fursin@cTuning.org, http://fursin.net
#

cfg={}  # Will be updated by CK (meta description of this module)
work={} # Will be updated by CK (temporal data)
ck=None # Will be updated by CK (initialized CK kernel) 

# Local settings
var_post_autorefresh='report_autorefresh'
var_post_autorefresh_time='report_autorefresh_time'

##############################################################################
# Initialize module

def init(i):
    """

    Input:  {}

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """
    return {'return':0}

##############################################################################
# viewing entry as html

def html_viewer(i):
    """
    Input:  {
              data_uoa

              url_base
              url_pull

              url_pull_tmp
              tmp_data_uoa

              url_wiki

              html_share

              form_name     - current form name

              (all_params)
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """
    import os

    h=''
    st=''
    raw='no'
    top='yes'

    duoa=i['data_uoa']
    burl=i['url_base']
    purl=i['url_pull']
    wurl=i.get('url_wiki','')

    tpurl=i['url_pull_tmp']
    tpuoa=i['tmp_data_uoa']

    ap=i.get('all_params',{})

    ruoa=ap.get('ck_top_repo','')
    muoa=ap.get('ck_top_module','')

    cparams=ap.get('graph_params','') # current graph params

    hshare=i.get('html_share','')

    form_name=i['form_name']
    form_submit='document.'+form_name+'.submit();'

    if duoa!='':
       # Load entry
       rx=ck.access({'action':'load',
                     'module_uoa':work['self_module_uid'],
                     'data_uoa':duoa})
       if rx['return']>0: return rx

       pp=rx['path']

       dd=rx['dict']
       duid=rx['data_uid']

       # Check what to do with top header
       if dd.get('top','')!='': top=dd['top']

       # Check auto-refresh
       dar=dd.get('auto_refresh','')
       if dar=='yes':
          ar=ap.get(var_post_autorefresh,'')
          art=ap.get(var_post_autorefresh_time,'')
          iart=4

          if art=='':
              art=dd.get('auto_refresh_time','')
          if art!='':
              try:
                 iart=int(art)
              except ValueError:
                 iart=4

          ap[var_post_autorefresh_time]=iart

          if ar=='':
              ar='on'
              ap[var_post_autorefresh]=ar

          if ar=='on':
             h+='\n'
             h+='<script language="javascript">\n'
             h+=' <!--\n'
             h+='  setTimeout(\''+form_submit+'\','+str(iart*1000)+');\n'
             h+=' //-->\n'
             h+='</script>\n'
             h+='\n'

       if dd.get('live','')!='yes':
          raw='yes'
       else:
          title=dd.get('title','')
          sub_title=dd.get('sub_title','')
          authors=dd.get('authors',[])
          affs=dd.get('affiliations',{})
          cauthor=dd.get('cor_author_email','')

          mwc=dd.get('media_wiki_commands','')

          h+='<div id="ck_entries">\n'

          h+='<center><span id="ck_article_title">'+title+'</span><br>\n'

          if sub_title!='':
             h+='<div id="ck_entries_space4"></div>\n'
             h+='<span id="ck_article_authors">\n'
             h+='<i>'+sub_title+'</i><br>\n'
             h+='</span>\n'

          if dd.get('skip_authors','')!='yes':
             if len(authors)!='':
                h+='<div id="ck_entries_space4"></div>\n'
                h+='<span id="ck_article_authors">\n'

                x=''
                for a in authors:
                    name=a.get('name','')
                    aff=a.get('affiliation','')

                    url=a.get('url','')
                    if url!='': name='<a href="'+url+'">'+name+'</a>'

                    if x!='': x+=', '
                    x+=name+'&nbsp;<sup>'+aff+'</sup>'

                h+=x+'<br>\n'
                h+='</span>\n'

             if len(affs)>0:
                h+='<div id="ck_entries_space4"></div>\n'
                h+='<span id="ck_article_affiliations">\n'

                x=''
                for a in sorted(affs, key=int):
                    af=affs[str(a)]
                    name=af.get('name','')

                    url=af.get('url','')
                    if url!='': name='<a href="'+url+'">'+name+'</a>'

                    if x!='': x+=',&nbsp;&nbsp;&nbsp;'
                    x+='<sup>'+str(a)+'</sup>&nbsp;'+name

                h+=x+'<br>\n'
                h+='</span>\n'

          if dd.get('add_date_to_the_top','')!='':
             h+='<span id="ck_article_affiliations">\n'
             h+='<div id="ck_entries_space4"></div>\n'
             h+='<i>\n'
             h+=dd['add_date_to_the_top']
             h+='</i>\n'
             h+='</span><br>\n'

          h+='</center>\n'

          if len(cauthor)>0:
             h+='<div id="ck_entries_space4"></div>\n'
             h+='<span id="ck_article_cauthor">\n'
             h+='<i>Corresponding author: <a href="mailto:'+cauthor+'">'+cauthor+'</a></i>\n'
             h+='</span>\n'

          if hshare!='' and dd.get('skip_sharing','')!='yes':
             h+='<div id="ck_entries_space4"></div>\n'
             h+=hshare
             h+=' <div id="ck_entries_space4"></div>\n'

          h+='<div style="text-align: right;">'

          x=''
          for q in dd.get('top_links',[]):
              qt=q['text']
              qu=q['url']

              if x!='': x+='&nbsp; '
              x+='[&nbsp;<a href="'+qu+'">'+qt+'</a>&nbsp;]\n'

          if x!='':
             h+='\n<br>\n'+x+'\n'

          top_urls=dd.get('top_urls',[])

          top_url=dd.get('top_url',{})
          if len(top_url)!=0:
             top_urls.append(top_url)

          for top_url in top_urls:
              n_top_url=top_url.get('name','')
              u_top_url=top_url.get('url','')
              if n_top_url!='' and u_top_url!='':
                 h+='[&nbsp;<a href="'+u_top_url+'">'+n_top_url+'</a>&nbsp;] '

          if wurl!='' and dd.get('skip_discussion_link','')!='yes':
             h+='[&nbsp;<a href="'+wurl+'">Discussion wiki (comments, reproducibility, etc.)</a>&nbsp;]'
          h+='</div>\n'

          h+='<div id="ck_entries_space4"></div>\n'
          h+='<div id="ck_entries_space4"></div>\n'

          # Checking template
          t=dd.get('template','')
          if t!='':
             h+='<span id="ck_article_text">\n'

             px=os.path.join(pp,t)
             th=''
             if os.path.isfile(px):
                rx=ck.load_text_file({'text_file':px})
                if rx['return']>0: return rx
                th=rx['string']

             # If MediaWiki, process extra
             if mwc=='yes':
                thx=th.split('\n')
                th=''

                for l in thx:
                    if l.startswith('=== '):
                       l='<h3>'+l[4:-4]+'</h3>'
                    elif l.startswith('== '):
                       l='<h2>'+l[3:-3]+'</h2>'
                    elif l.startswith('= '):
                       l='<h1>'+l[2:-2]+'</h1>'

                    j=l.find('[')
                    while j>=0:
                       j1=l.find(']',j)
                       if j1>0:
                          url=l[j+1:j1]
                          txt=url
                          j2=url.find(' ')
                          if j2>0:
                             txt=url[j2+1:].strip()
                             url=url[:j2].strip()

                          l=l[:j]+'<a href="'+url+'">'+txt+'</a>'+l[j1+1:]
                          
                       j=l.find('[',j+1)

                    if l=='': l='<p>'

                    th+=l
#                    if not l.endswith('>'):
#                       th+='<br>'
                    th+='\n'

             # Check CK access functions
             h+=th

             j=0
             while j!=-1:
                j=h.find('$#ck_access_start#$',j)
                if j>=0:
                   j1=h.find('$#ck_access_stop#$',j)
                   if j1>=0:
                       json=h[j+19:j1].strip()

                       ha=''

                       # Convert json to python dict
                       rx=ck.convert_json_str_to_dict({'str':json, 'skip_quote_replacement':'yes'})
                       if rx['return']==0:
                          nii=rx['dict']

                          if nii.get('action','')!='':
                             # Here we process active CK plugin (if there is an "action") to generate content (even possibly from remote repo)
                             nii['base_url']=burl

                             rx=ck.access(nii)
                             if rx['return']==0:
                                 ha=rx.get('html','')

                                 if nii.get('remove_script_src','')=='yes':
                                    k1=ha.find('<script src')
                                    if k1>=0:
                                       k2=ha.find('</script>', k1)
                                       if k2>0:
                                          ha=ha[:k1]+ha[k2+9:]

                                 st+=rx.get('style','')
                             else:
                                 ha+='\n<br>\nInternal error: '+rx['error']+'\n<br>\n'

                       h=h[:j]+ha+h[j1+18:]
                       
                   j+=1

             h+='</span\n'
          h+='</div>\n'

          # Checking global style
          xt=dd.get('style','')
          if t!='':
             px=os.path.join(pp,xt)
             if os.path.isfile(px):
                rx=ck.load_text_file({'text_file':px})
                if rx['return']>0: return rx
                st+=rx['string']
          st=st.replace('$#ck_root_url#$', burl)

          if dar=='yes':
             h+='<center>\n'
             checked=''
             if ar=='on': checked=' checked '
             h+='&nbsp;&nbsp;&nbsp;Auto-replot graph:&nbsp;<input type="checkbox" name="'+var_post_autorefresh+'" id="'+var_post_autorefresh+'" onchange="submit()"'+checked+'>,'
             h+='&nbsp;seconds: <input type="text" name="'+var_post_autorefresh_time+'" value="'+str(iart)+'">\n'
             h+='</center>\n'

          if dd.get('add_ck_info','')!='':
             h+='<br>\n'
             h+=' <center><i>\n'
             h+='  This interactive and reproducible article was automatically generated using <b><a href="http://github.com/ctuning/ck">Collective Knowledge framework</a></b>'
             h+=' </i></center><br>\n'

    return {'return':0, 'raw':raw, 'show_top':top, 'html':h, 'style':st}

##############################################################################
# copy file from an entry to current directory

def copy_file(i):
    """
    Input:  {
              (module_uoa) - module with file
              data_uoa     - entry with file
              file         - file to copy
              dir          - extra dir in current directory
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    import os
    import filecmp
    import shutil

    muoa=i['module_uoa']
    duoa=i['data_uoa']
    f=i['file']

    r=ck.access({'action':'find',
                 'module_uoa':muoa,
                 'data_uoa':duoa})
    if r['return']>0: return r

    p=r['path']

    pp=os.path.join(p,f)

    if not os.path.isfile(pp):
       return {'return':1, 'error':'file not found ('+pp+')'}

    ppn=os.getcwd()
    if i.get('dir','')!='':
       ppn=os.path.join(ppn,i['dir'])

    ppd=ppn

    ppn=os.path.join(ppn,f)

    if not os.path.isfile(ppn) or not filecmp.cmp(pp,ppn):
       if not os.path.isdir(ppd):
          os.makedirs(ppd)
    
       ck.out('Copying file '+f+' to '+ppn+' ...')
       shutil.copyfile(pp,ppn)

    return {'return':0}



#
# Collective Knowledge (processing XML)
#
# See CK LICENSE.txt for licensing details
# See CK COPYRIGHT.txt for copyright details
#
# Developer: Grigori Fursin, Grigori.Fursin@cTuning.org, http://fursin.net
#

cfg={}  # Will be updated by CK (meta description of this module)
work={} # Will be updated by CK (temporal data)
ck=None # Will be updated by CK (initialized CK kernel) 

# Local settings

##############################################################################
# Initialize module

def init(i):
    """

    Input:  {}

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """
    return {'return':0}

##############################################################################
# validate XML against DTD

def validate(i):
    """
    Input:  {
              (data_uoa) - if specified use DTD file from that entry
              (dtd_module_uoa) - if specified use DTD file from that entry, otherwise 'xml'
              dtd_file   - DTD file
              xml_file   - XML file to be validated
            }

    Output: {
              return       - return code =  0, if successfully validated
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    import os
    from lxml import etree

    ocon=i.get('out','')=='con'

    # Checking entry with DTD if needed
    p=''
    duoa=i.get('data_uoa','')
    dd={}
    if duoa!='':
       muoa=i.get('dtd_module_uoa','')
       if muoa=='': muoa=i.get('module_uoa','')

       r=ck.access({'action':'load',
                    'module_uoa':muoa,
                    'data_uoa':duoa})
       if r['return']>0: return r

       p=r['path']
       dd=r['dict']

    # Checking DTD file
    dtd_file=i.get('dtd_file','')
    if dtd_file=='': 
       dtd_file=dd.get('dtd_file','')
    if dtd_file=='': 
       return {'return':1, 'error':'"dtd_file" is not specified'}

    if p!='':
       dtd_file=os.path.join(p, dtd_file)

    if not os.path.isfile(dtd_file):
       return {'return':1, 'error':'DTD file not found ('+dtd_file+')'}

    xml_file=i.get('xml_file','')
    if xml_file=='': 
       return {'return':1, 'error':'"xml_file" is not specified'}

    if not os.path.isfile(xml_file):
       return {'return':1, 'error':'XML file not found ('+xml_file+')'}

    # Load DTD
    fdtd=open(dtd_file)
    DTD=etree.DTD(fdtd)

    # Load XML
    xml=etree.parse(xml_file)

    # Validate
    validated=DTD.validate(xml)
 
    fdtd.close()

    if not validated:
       return {'return':1, 'error':'XML was not validated:\n'+str(DTD.error_log.filter_from_errors())}

    if ocon:
       ck.out('XML file was successfully validated against DTD!')

    return {'return':0}


##############################################################################
# generate XML element
def generate_element(root, d):

    import os
    from lxml import etree

    if type(d)!=list:
       return {'return':1, 'error':'unknown format - expected list but got "'+str(type(d))+'"'}

    for l in d:
        if type(l)!=dict:
           return {'return':1, 'error':'unknown format - expected dict in a list but got "'+str(type(d))+'"'}

        if len(l)>1:
           return {'return':1, 'error':'unknown format - dict must have size 1 but got "'+str(len(l))+'"'}

        k=list(l.keys())[0]

        e=etree.Element(k)

        v=l[k]

        if type(v)==list:
           r=generate_element(e,v)
           if r['return']>0: return r
        else:
           v=str(v)
           if v!='':
              # Check $% %$
              if v.startswith('$%'):
                 j=v.find('%$')
                 if j>0:
                    v1=v[2:j].strip()
                    v=v[j+2:].strip()

                    # Convert JSON to dict (attributes)
                    r=ck.convert_json_str_to_dict({'str':v1, 'skip_quote_replacement':'yes'})
                    if r['return']>0: return r
                    attributes=r['dict']

                    e=etree.Element(k, attrib=attributes)

              e.text=v

        root.append(e)

    return {'return':0}

##############################################################################
# generate XML from DICT (JSON)

def generate(i):
    """
    Input:  {
              (json_file) - JSON file with dict in a specific format (starts with list, all leafs are also lists or values)
                 or
              (dict)      - dict in a specific format (starts with list, all leafs are also lists or values)

              (root_key)           - root name (such as "proceedings")
              (root_attributes)    - root attributes

              xml_file    - file to save XML

              (dtd_file)  - DTD file to validate produced XML

              if dtd_file!='':
                (data_uoa)       - if specified use DTD file from that entry
                (dtd_module_uoa) - if specified use DTD file from that entry, otherwise 'xml'

            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    import os
    from lxml import etree

    ocon=i.get('out','')=='con'

    xml_file=i.get('xml_file','')
    if xml_file=='':
       return {'return':1, 'error':'"xml_file" is not defined'}

    # Get dictionary
    jf=i.get('json_file','')
    if jf!='':
       r=ck.load_json_file({'json_file':jf})
       if r['return']>0: return r

       d=r['dict']
    else:
       d=i.get('dict',[])

    # Prepare XML root
    root_key=i.get('root_key','')
    if root_key=='': root_key='root'

    attributes=i.get('root_attributes',{})
#    root = etree.Element('proceeding', ts='05/09/2017', ver='10.1')

    root = etree.Element(root_key, attrib=attributes)

    # Prepare XML doc (recursively)
    if ocon: ck.out('* Generating XML ...')
    r=generate_element(root,d)
    if r['return']>0: return r

#    root.append(r['element'])

    # Save to XML file
    if ocon: ck.out('* Recording XML ...')
    xml = etree.ElementTree(root)
    xml.write(xml_file, xml_declaration=True, encoding='UTF-8', pretty_print=True) 

    # Validating XML file
    dtd_file=i.get('dtd_file','')
    if dtd_file!='':
       if ocon: ck.out('* Validating XML ...')

       duoa=i.get('data_uoa','')
       muoa=i.get('dtd_module_uoa','')
       if muoa=='':
          muoa=work['self_module_uid']

       r=validate({'data_uoa':duoa,
                   'dtd_module_uoa':muoa,
                   'dtd_file':dtd_file,
                   'xml_file':xml_file})
       if r['return']>0: return r

    return {'return':0}



#
# Collective Knowledge - common functionality for MLPerf benchmarking.
#
# See CK LICENSE.txt for licensing details.
# See CK COPYRIGHT.txt for copyright details.
#
# Developers:
# - Leo Gordon, leo@dividiti.com
# - Anton Lokhmotov, anton@dividiti.com
#

cfg={}  # Will be updated by CK (meta description of this module)
work={} # Will be updated by CK (temporal data)
ck=None # Will be updated by CK (initialized CK kernel)

import getpass
import os
import sys
from pprint import pprint


def init(i):
    """
    Not to be called directly.
    """

    return {'return':0}


def compare_experiments_image_classification(frame_predictions):
    epsilon = 1e-5 # 1/1000th of a percent (1e-3 * 1e-2)
    max_delta = 0
    num_mismatched_files = 0
    num_mismatched_classes = 0
    num_mismatched_probabilities = 0
    num_mismatched_elementary_keys = 0
    for file_name in sorted(frame_predictions[0]):
        ck.out( 'Checking {}...'.format(file_name) )
        fp0 = frame_predictions[0][file_name]
        fp1 = frame_predictions[1][file_name]
        # Check elementary keys.
        for key in [ 'accuracy_top1', 'accuracy_top5', 'accuracy_topn', 'class_correct', 'class_topmost' ]:
            fp0_value = fp0[key]
            fp1_value = fp1[key]
            if fp0_value != fp1_value:
                ck.out( '- mismatched key {}: {} != {}'.format(key, fp0_value, fp1_value) )
                num_mismatched_elementary_keys += 1
        # Check topn.
        index = 0
        any_mismatched_classes = False
        any_mismatched_probabilities = False
        for (fp0_topn, fp1_topn) in zip(fp0['topn'], fp1['topn']):
            # Compare classes.
            if fp0_topn['index'] != fp1_topn['index']:
                ck.out( '- mismatched classes at index {}: {} != {}'.format(index, fp0_topn['index'], fp1_topn['index']) )
                num_mismatched_classes += 1
                any_mismatched_classes = True
            # Compare probabilities.
            delta = abs(fp0_topn['prob'] - fp1_topn['prob'])
            if delta > epsilon:
                ck.out( '- mismatched probabilities at index {}: | {:.5f} - {:.5f} | = {:.5f} > {:.5f}'.format(index, fp0_topn['prob'], fp1_topn['prob'], delta, epsilon) )
                num_mismatched_probabilities += 1
                any_mismatched_probabilities = True
            if delta > max_delta: max_delta = delta
            index += 1
        if any_mismatched_classes or any_mismatched_probabilities:
            num_mismatched_files += 1

    rdict = { 'return':0,
              'epsilon':epsilon,
              'max_delta':max_delta,
              'num_mismatched_files':num_mismatched_files,
              'num_mismatched_classes':num_mismatched_classes,
              'num_mismatched_probabilities':num_mismatched_probabilities,
              'num_mismatched_elementary_keys':num_mismatched_elementary_keys
    }

    return rdict


def compare_experiments_object_detection(frame_predictions):
    # Probabilities.
    epsilon_prob   = 1e-5 # 1/1000th of a percent (1e-3 * 1e-2)
    max_delta_prob = 0.0
    num_mismatched_probabilities = 0
    # Bounding boxes.
    epsilon_bbox   = [1.0, 1.0, 1.0, 1.0] # 1 pixel in any direction
    max_delta_bbox = [0.0, 0.0, 0.0, 0.0]
    num_mismatched_bboxes = 0
    # Squared distances of the centre of mass.
    epsilon_dist   = 100.0 # 10.0 squared
    max_delta_dist = 0.0
    num_mismatched_distances = 0
    # Files and classes.
    num_mismatched_files = 0
    num_mismatched_classes = 0

    for file_name in sorted(frame_predictions[0]):
        ck.out( 'Checking {}...'.format(file_name) )
        fp0 = frame_predictions[0][file_name]
        fp1 = frame_predictions[1][file_name] 
        if len(fp0['detections']) == len(fp1['detections']):
            mismatch_detected_objects_count = False
        else:
            ck.out( '- mismatch detected objects count: {} != {}'.format(len(fp0['detections']), len(fp1['detections'])) )
            mismatch_detected_objects_count = True

        index = 0
        any_mismatched_classes = False
        any_mismatched_probabilities = False
        any_mismatched_bbox = False
        for (fpa, fpb) in zip(fp0['detections'], fp1['detections']):
            # Extract the coordinates.
            fpa_x1 = fpa['bbox'][0]
            fpa_y1 = fpa['bbox'][1]
            fpa_x2 = fpa['bbox'][2]
            fpa_y2 = fpa['bbox'][3]
            fpb_x1 = fpb['bbox'][0]
            fpb_y1 = fpb['bbox'][1]
            fpb_x2 = fpb['bbox'][2]
            fpb_y2 = fpb['bbox'][3]
            # Compute the distance to the centre of mass.
            fpa_dist = ((fpa_x2-fpa_x1)*(fpa_x2-fpa_x1) + (fpa_y2-fpa_y1)*(fpa_y2-fpa_y1))/4
            fpb_dist = ((fpb_x2-fpb_x1)*(fpb_x2-fpb_x1) + (fpb_y2-fpb_y1)*(fpb_y2-fpb_y1))/4
            # Compare the distances.
            delta_dist = abs(fpa_dist - fpb_dist)
            if delta_dist > epsilon_dist:
                ck.out( '- mismatched distances at index {}: [{}] != [{}]'.format(index, fpa_dist, fpb_dist) )
                num_mismatched_distances += 1
            if delta_dist > max_delta_dist: max_delta_dist = delta_dist
            # Compare classes.
            if fpa['class'] != fpb['class']:
                ck.out( '- mismatched classes at index {}: [{}] != [{}]'.format(index, fpa['class'], fpb['class']) )
                num_mismatched_classes += 1
                any_mismatched_classes = True
            # Compare probabilities.
            delta_prob = abs(fpa['prob'] - fpb['prob'])
            if delta_prob > epsilon_prob:
                ck.out( '- mismatched probabilities at index {}: | {:.5f} - {:.5f} | = {:.5f} > {:.5f}'.format(index, fpa['prob'], fpb['prob'], delta_prob, epsilon_prob) )
                num_mismatched_probabilities += 1
                any_mismatched_probabilities = True
            if delta_prob > max_delta_prob: max_delta_prob = delta_prob
            # Compare coordinates.
            delta_x1 = abs(fpa_x1 - fpb_x1)
            delta_y1 = abs(fpa_y1 - fpb_y1)
            delta_x2 = abs(fpa_x2 - fpb_x2)
            delta_y2 = abs(fpa_y2 - fpb_y2)
            if delta_x1 > epsilon_bbox[0] or delta_y1 > epsilon_bbox[1] or delta_x2 > epsilon_bbox[2] or delta_y2 > epsilon_bbox[3]:
                ck.out( '- mismatched bbox at index {}: [{}] != [{}]'.format(index, fpa['bbox'], fpb['bbox']) )
                any_mismatched_bbox = True
                num_mismatched_bboxes += 1
            if delta_x1 > max_delta_bbox[0]: max_delta_bbox[0] = delta_x1
            if delta_y1 > max_delta_bbox[1]: max_delta_bbox[1] = delta_y1
            if delta_x2 > max_delta_bbox[2]: max_delta_bbox[2] = delta_x2
            if delta_y2 > max_delta_bbox[3]: max_delta_bbox[3] = delta_y2
            index += 1

        if mismatch_detected_objects_count or any_mismatched_probabilities or any_mismatched_bbox:
            num_mismatched_files +=1

    rdict = { 'return':0,
              'epsilon_prob':epsilon_prob,
              'max_delta_prob':max_delta_prob,
              'epsilon_bbox':epsilon_bbox,
              'max_delta_bbox':max_delta_bbox,
              'epsilon_dist':epsilon_dist,
              'max_delta_dist':max_delta_dist,
              'num_mismatched_files':num_mismatched_files,
              'num_mismatched_classes':num_mismatched_classes,
              'num_mismatched_distances':num_mismatched_distances,
              'num_mismatched_probabilities':num_mismatched_probabilities
    }

    return rdict


def compare_experiments(i):
    """
    Input:  {
                (cids[])            - up to 2 CIDs of entries to compare (interactive by default)
                (repo_uoa)          - experiment repository ('*' by default)
                (extra_tags)        - extra tags to search for CIDs
            }

    Output: {
                return              - return code =  0, if successful
                                                  >  0, if error
                (error)             - error text if return > 0
            }
    """
    cids = i.get('cids')
    repo_uoa = i.get('repo_uoa', '*')
    extra_tags = i.get('extra_tags', 'mlperf,accuracy')

    # Return an error if more than 2 CIDs have been provided.
    if len(cids) > 2:
        return {'return':1, 'error': 'only support up to 2 CIDs'}

    # Interactively select experiment entries if fewer than 2 CID have been provided.
    for i in range(len(cids),2):
        ck.out( 'Select experiment #{} for comparison:'.format(i) )
        pick_exp_adict = { 'action':       'pick_an_experiment',
                           'module_uoa':   'mlperf',
                           'repo_uoa':     repo_uoa,
                           'extra_tags':   extra_tags,
        }
        r=ck.access( pick_exp_adict )
        if r['return']>0: return r
        cids.append( r['cid'] )

    # Collect frame predictions.
    ck.out( '\nExperiments to compare:' )
    experiment_types  = []
    frame_predictions = []
    accuracy_top1     = []
    accuracy_top5     = []
    accuracy_mAP      = []
    accuracy_recall   = []
    for cid in cids:
        r=ck.parse_cid({'cid': cid})
        if r['return']>0:
            return { 'return': 1, 'error': "Cannot parse CID '{}'".format(cid) }
        else:
            repo_uoa    = r.get('repo_uoa','')
            module_uoa  = r.get('module_uoa','')
            data_uoa    = r.get('data_uoa','')

        # Load experimental point.
        load_point_adict = {    'action':           'load_point',
                                'repo_uoa':         repo_uoa,
                                'module_uoa':       module_uoa,
                                'data_uoa':         data_uoa,
        }
        r=ck.access( load_point_adict )
        if r['return']>0: return r

        point0001_characteristics_list = r['dict']['0001']['characteristics_list']
        point0001_characteristics0_run = point0001_characteristics_list[0]['run']
        point0001_frame_predictions    = point0001_characteristics0_run['frame_predictions']
        ck.out( '- {}: {} predictions'.format(cid, len(point0001_frame_predictions)) )
        frame_predictions.append(point0001_frame_predictions)

        # Load pipeline to determine experiment type from tags.
        load_pipeline_adict = { 'action':           'load_pipeline',
                                'repo_uoa':         repo_uoa,
                                'module_uoa':       module_uoa,
                                'data_uoa':         data_uoa,
        }
        r=ck.access( load_pipeline_adict )
        if r['return']>0: return r

        tags = r['pipeline']['tags'].split(',')
        if 'image-classification' in tags:
            experiment_types.append('image-classification')
            accuracy_top1.append(point0001_characteristics0_run['accuracy_top1'])
            accuracy_top5.append(point0001_characteristics0_run['accuracy_top5'])
        if 'object-detection' in tags:
            experiment_types.append('object-detection')
            accuracy_mAP.append(point0001_characteristics0_run['mAP'])
            accuracy_recall.append(point0001_characteristics0_run['recall'])

    unique_experiment_types = list(set(experiment_types))
    if len(unique_experiment_types)!=1:
        return { 'return': 2, 'error': "Mixed experiment types: '{}'".format(str(unique_experiment_types)) }
    experiment_type = unique_experiment_types[0]
    ck.out('Experiment type: {}'.format(experiment_type))
    if experiment_type=='image-classification':
        rdict = compare_experiments_image_classification(frame_predictions)
        rdict['delta_top1']   = accuracy_top1[0] - accuracy_top1[1]
        rdict['delta_top5']   = accuracy_top5[0] - accuracy_top1[1]
    elif experiment_type=='object-detection':
        rdict = compare_experiments_object_detection(frame_predictions)
        rdict['delta_mAP']    = accuracy_mAP[0] - accuracy_mAP[1]
        rdict['delta_recall'] = accuracy_recall[0] - accuracy_recall[1]
    else:
        return { 'return': 3, 'error': "Unknown experiment type: '{}'".format(experiment_type) }

    pprint( rdict )

    return rdict


def list_experiments(i):
    """
    Input:  {
                (repo_uoa)          - experiment repository name ('*' by default)
                (extra_tags)        - extra tags to filter
                (add_meta)          - request to return metadata with each experiment entry
            }

    Output: {
                return              - return code =  0, if successful
                                                  >  0, if error
                (error)             - error text if return > 0
            }
    """

    repo_uoa        = i.get('repo_uoa', '*')
    extra_tags      = i.get('extra_tags')
    all_tags        = 'mlperf' + ( ',' + extra_tags if extra_tags else '' )
    add_meta        = i.get('add_meta')

    search_adict    = { 'action':       'search',
                        'repo_uoa':     repo_uoa,
                        'module_uoa':   'experiment',
                        'data_uoa':     '*',
                        'tags':         all_tags,
                        'add_meta':     add_meta,
    }
    r=ck.access( search_adict )
    if r['return']>0: return r

    list_of_experiments = r['lst']

    repo_to_names_list = {}
    for entry in list_of_experiments:
        repo_uoa    = entry['repo_uoa']
        data_uoa    = entry['data_uoa']
        if not repo_uoa in repo_to_names_list:
            repo_to_names_list[ repo_uoa ] = []
        repo_to_names_list[ repo_uoa ] += [ data_uoa ]

    if i.get('out')=='con':
        for repo_uoa in repo_to_names_list:
            experiments_this_repo = repo_to_names_list[ repo_uoa ]
            ck.out( '{} ({}) :'.format(repo_uoa, len(experiments_this_repo) ) )
            for data_uoa in experiments_this_repo:
                ck.out( '\t' + data_uoa )
            ck.out( '' )

    return {'return':0, 'lst': list_of_experiments, 'repo_to_names_list': repo_to_names_list}


def pick_an_experiment(i):
    """
    Input:  {
                (repo_uoa)          - experiment repository ('*' by default)
                (extra_tags)        - extra tags to filter
            }

    Output: {
                return              - return code =  0, if successful
                                                  >  0, if error
                (error)             - error text if return > 0
            }
    """

    repo_uoa        = i.get('repo_uoa', '*')
    extra_tags      = i.get('extra_tags')

    list_exp_adict  = { 'action':       'list_experiments',
                        'module_uoa':   'mlperf',
                        'repo_uoa':     repo_uoa,
                        'extra_tags':   extra_tags,
    }
    r=ck.access( list_exp_adict )
    if r['return']>0: return r

    if len(r['lst'])==0:
        return {'return':1, 'error':'No experiments to choose from - please relax your filters'}

    all_experiment_names = [ '{repo_uoa}:{module_uoa}:{data_uoa}'.format(**entry_dict) for entry_dict in r['lst']]

    select_adict = {'action': 'select_string',
                    'module_uoa': 'misc',
                    'options': all_experiment_names,
                    'default': '0',
                    'question': 'Please select one of the experiment entries',
    }
    r=ck.access( select_adict )
    if r['return']>0:
        return r
    else:
        cid = r['selected_value']

    return {'return':0, 'cid': cid}




#
# Collective Knowledge (os)
#
# See CK LICENSE.txt for licensing details
# See CK COPYRIGHT.txt for copyright details
#
# Developer: Grigori Fursin, Grigori.Fursin@cTuning.org, http://fursin.net
#

cfg={}  # Will be updated by CK (meta description of this module)
work={} # Will be updated by CK (temporal data)
ck=None # Will be updated by CK (initialized CK kernel) 

# Local settings

##############################################################################
# Initialize module

def init(i):
    """

    Input:  {}

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """
    return {'return':0}

##############################################################################
# find close OS

def find_close(i):
    """
    Input:  {
              (os_uoa)     - load info from a given OS
              (os_dict)    - if non-empty, return this dict
            }

    Output: {
              return     - return code =  0

              platform   - 'win' or 'linux'. Careful - it is always for current host OS! 
                           Use 'ck_name' key from meta for the target OS!

              bits       - (str) 32 or 64. Careful - it is always for current host OS!
                           Use 'bits' key from meta for the target OS!

              os_uoa     - UOA of the most close OS
              os_uid     - UID of the most close OS
              os_dict    - meta of the most close OS

              (add_path) - list of extra path ...
            }
    """

    r=ck.get_os_ck({})
    if r['return']>0: return r

    bits=r['bits']
    plat=r['platform']

    xos=i.get('os_uoa','')
    fc=i.get('find_close','')

    if xos=='':
       # Detect host platform
       # Search the most close OS
       ii={'action':'search',
           'module_uoa':work['self_module_uid'],
           'search_dict':{'ck_name':plat,
                          'bits':bits,
                          'generic':'yes',
                          'priority':'yes'},
           'internal':'yes'}

       # Adding extra tags to separate different Linux flavours such as Mac OS X:
       import sys
       pl=sys.platform

       if pl=='darwin':
          ii['tags']='macos'
       elif plat=='linux':
          ii['tags']='standard'

       rx=ck.access(ii)
       if rx['return']>0: return rx

       lst=rx['lst']
       if len(lst)==0:
          return {'return':1, 'error':'most close platform was not found in CK'}

       pl=lst[0]

       xos=pl.get('data_uoa','')

    rr={'return':0, 'platform':plat, 'bits':bits}

    # Load OS
    if xos!='':
       r=ck.access({'action':'load',
                    'module_uoa':'os', 
                    'data_uoa':xos})
       if r['return']>0: return r

       os_uoa=r['data_uoa']
       os_uid=r['data_uid']

       dd=r['dict']

       if len(i.get('os_dict',{}))!=0: # Substitute from 'machine' description (useful for remote access)
           dd=i['os_dict']

       rr['os_uoa']=os_uoa
       rr['os_uid']=os_uid
       rr['os_dict']=dd

       # Check if need to add path
       x=dd.get('add_to_path_os_uoa','')
       if x!='':
          rx=ck.access({'action':'find',
                        'module_uoa':work['self_module_uid'],
                        'data_uoa':x})
          if rx['return']>0: return rx
          px=rx['path']

          rr['add_path']=[px]

    return rr

##############################################################################
# shell in OS

def shell(i):
    """
    Input:  {
              (target)
              (host_os)
              (target_os)
              (device_id)

              (cmd) -               cmd string (can have \n)

              (split_to_list) -     if 'yes', split stdout and stderr to list

              (should_be_remote) -  if 'yes', can run only on remote target

              (output_to_console) - if 'yes', output to console instead of files

              (encoding)            - if !='', use this encoding
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0

              return_code

              stdout or stdout_lst
              stderr or stderr_lst
            }

    """

    import os

    o=i.get('out','')
    oo=''
    if o=='con': oo='con'

    stl=i.get('split_to_list','')
    sbr=i.get('should_be_remote','')
    otc=i.get('output_to_console','')

    # Check if need to initialize device and directly update input i !
    ii={'action':'init',
        'module_uoa':cfg['module_deps']['machine'],
        'input':i}
    r=ck.access(ii)
    if r['return']>0: return r

    device_cfg=i.get('device_cfg',{})

    encoding=i.get('encoding','')

    # Check host/target OS/CPU
    hos=i.get('host_os','')
    tos=i.get('target_os','')
    tdid=i.get('device_id','')

    # Get some info about platforms
    ii={'action':'detect',
        'module_uoa':cfg['module_deps']['platform.os'],
        'host_os':hos,
        'target_os':tos,
        'device_cfg':device_cfg,
        'device_id':tdid,
        'skip_info_collection':'yes'}
    r=ck.access(ii)
    if r['return']>0: return r

    hos=r['host_os_uid']
    hosx=r['host_os_uoa']
    hosd=r['host_os_dict']

    tos=r['os_uid']
    tosx=r['os_uoa']
    tosd=r['os_dict']

    tdid=i.get('device_id','')

    xtdid=''
    if tdid!='': xtdid=' -s '+tdid

    envsep=hosd.get('env_separator','')
    sext=hosd.get('script_ext','')
    sexe=hosd.get('set_executable','')
    sbp=hosd.get('bin_prefix','')
    scall=hosd.get('env_call','')
    ubtr=hosd.get('use_bash_to_run','')
    stro=tosd.get('redirect_stdout','')
    stre=tosd.get('redirect_stderr','')

    cmd=i.get('cmd','')

    if otc!='yes':
       # Tmp file for stdout
       rx=ck.gen_tmp_file({'prefix':'tmp-', 'suffix':'.tmp', 'remove_dir':'no'})
       if rx['return']>0: return rx
       fno=rx['file_name']

       # Tmp file for stderr
       rx=ck.gen_tmp_file({'prefix':'tmp-', 'suffix':'.tmp', 'remove_dir':'no'})
       if rx['return']>0: return rx
       fne=rx['file_name']

    # Check remote shell
    rs=tosd.get('remote_shell','')
    if sbr=='yes' and rs=='':
        return {'return':1, 'error':'target mush be remote'}

    if rs!='':
        # ADB dependency
        deps={'adb':{
                     "force_target_as_host": "yes",
                     "local": "yes", 
                     "name": "adb tool", 
                     "sort": -10, 
                     "tags": "tool,adb"
                     }
             }

        ii={'action':'resolve',
            'module_uoa':cfg['module_deps']['env'],
            'host_os':hos,
            'target_os':tos,
            'device_id':tdid,
            'deps':deps,
            'add_customize':'yes',
            'out':oo}
        rx=ck.access(ii)
        if rx['return']>0: return rx

        x='"'
        cmd=rx['cut_bat']+'\n'+rs.replace('$#device#$',xtdid)+' '+x+cmd+x

    # Record to tmp batch and run
    rx=ck.gen_tmp_file({'prefix':'tmp-', 'suffix':sext, 'remove_dir':'no'})
    if rx['return']>0: return rx
    fn=rx['file_name']

    rx=ck.save_text_file({'text_file':fn, 'string':cmd})
    if rx['return']>0: return rx

    # Prepare CMD for the host
    y=''
    if sexe!='':
       y+=sexe+' '+fn+envsep
    y+=' '+scall+' '+fn

    if ubtr!='': y=ubtr.replace('$#cmd#$',y)

    if otc!='yes':
       y=y+' '+stro+' '+fno+' '+stre+' '+fne

    rx=os.system(y)

    if os.path.isfile(fn):
        os.remove(fn)

    rr={'return':0, 'return_code':rx, 'target_os_dict':tosd}

    # Reading stdout file
    if otc!='yes':
       rx=ck.load_text_file({'text_file':fno, 'delete_after_read':'yes', 'split_to_list':stl, 'encoding':encoding})
       if rx['return']>0: return rx

       if stl=='yes':
           stdout=''

           rr['stdout_lst']=rx['lst']

           for q in rx['lst']:
               stdout+=q+'\n'
       else:
           stdout=rx['string']
           rr['stdout']=stdout

       # Reading stderr file
       rx=ck.load_text_file({'text_file':fne, 'delete_after_read':'yes', 'split_to_list':stl, 'encoding':encoding})
       if rx['return']>0: return rx
       if stl=='yes':
           stderr=''

           rr['stderr_lst']=rx['lst']

           for q in rx['lst']:
               stderr+=q+'\n'
       else:
           stderr=rx['string']
           rr['stderr']=stderr

       # Print if needed
       if o=='con':
          if stdout!='': ck.out(stdout)
          if stderr!='': ck.eout(stderr)

    return rr

##############################################################################
# convert UID to alias (UID) for user-friendly printing

def convert_uid_to_alias(i):
    """
    Input:  {
              uoa - OS uoa
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
              string - "UID" if now alias or "alias (UID)"
            }

    """

    uoa=i['uoa']

    s=uoa

    r=ck.access({'action':'load',
                 'module_uoa':work['self_module_uid'],
                 'data_uoa':uoa})
    if r['return']>0: return r

    da=r['data_alias']
    if da!='':
       s=da+' ('+r['data_uid']+')'

    return {'return':0, 'string':s}

##############################################################################
# generates shell script for exporting library path variables for the given platform

def lib_path_export_script(i):
    """
    Input:  {
              host_os_dict          - host OS meta
              (dynamic_lib_path)    - dynamic (shared) library path, or a list of paths
              (static_lib_path)     - dynamic (shared) library path, or a list of paths
              (lib_path)            - if set, and dynamic_path/static_lib_path is not set, 
                                      uses this value as dynamic_lib_path and/or static_lib_path
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0

              script       - script to execute to add the given path to the dynamic and/or static library path env vars
            }

    """

    host_d = i.get('host_os_dict', {})
    dynamic_path = _convert_lib_path_list_to_string(i.get('dynamic_lib_path', i.get('lib_path', '')))
    static_path = _convert_lib_path_list_to_string(i.get('static_lib_path', i.get('lib_path', '')))

    dynamic_var_name = host_d.get('env_ld_library_path', 'LD_LIBRARY_PATH')
    static_var_name = host_d.get('env_library_path', 'LIBRARY_PATH')

    hplat=host_d.get('ck_name','')
    s = ''
    if hplat != 'win' and (dynamic_path != '' or static_path != ''):
      s += '\n'
      if dynamic_path != '':
        s += 'export {name}="{value}":${name}\n'.format(name=dynamic_var_name, value=dynamic_path)
      if static_path != '':
        s += 'export {name}="{value}":${name}\n'.format(name=static_var_name, value=static_path)
      s += '\n'

    return {'return': 0, 'script': s}

##############################################################################
def _convert_lib_path_list_to_string(lst):
    return '":"'.join(lst) if isinstance(lst, list) else lst

##############################################################################
# convert win path to cygwin path

def convert_to_cygwin_path(i):
    """
    Input:  {
              path - path to convert
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0

              path         - updated path
            }

    """

    p=i['path']

    pp=p.replace('\\','/')

    if i.get('out','')=='con':
       ck.out(pp)

    return {'return':0, 'path':pp}

##############################################################################
# convert multiple paths to cygwin paths

def convert_to_cygwin_paths(i):
    """
    Input:  {
              paths - dict with paths to convert
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0

              paths - updated dict with paths
            }

    """

    pp=i['paths']

    for k in pp:
        p=pp[k]

        r=convert_to_cygwin_path({'path':p})
        if r['return']>0: return r

        pp[k]=r['path']

    return {'return':0, 'paths':pp}

##############################################################################
# run command and get stdout
def run_and_get_stdout(i):
  """
  Input:  {
            cmd       - list of command line arguments, starting with the command itself
            (shell)   - if 'yes', reuse shell environment
          }

  Output: {
            return       - return code =  0, if successful
                                       >  0, if error
                                       =  8, if timeout
            (error)      - error text if return > 0

            return_code  - return code from app

            stdout       - string, standard output of the command
          }
  """

  import subprocess
  import shlex
  import platform
  import sys

  cmd=i['cmd']
  if type(cmd)!=list:
      # Split only on non-Windows platforms (since Windows takes a string in Popen)
      if not platform.system().lower().startswith('win'):
          cmd=shlex.split(cmd)

  xshell=False
  if i.get('shell','')=='yes':
      xshell=True

  p1 = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=xshell)
  output, error = p1.communicate()

  if sys.version_info[0]>2:
      output = output.decode(encoding='UTF-8')
      error = error.decode(encoding='UTF-8')

  return {'return':0, 'return_code':p1.returncode, 'stdout':output, 'stderr':error}

##############################################################################
# find file or directory in all above directories from a given path

def find_file_above_path(i):
    """
    Input:  {
              (path) - if not specified, use current path
              file   - search for file or directory
            }

    Output: {
              path   - path where file or directory was found (or empty if not found)

              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    import os

    path_found=''

    sfile=i['file']

    path=i.get('path','')
    if path=='':
       path=os.getcwd()

    while True:
       path1=os.path.dirname(path)

       print (path1)

       if path1==path:
          break

       path=path1

       path2=os.path.join(path, sfile)
       print (path2)
       if os.path.isfile(path2) or os.path.isdir(path2):
          path_found=path
          break

    return {'return':0, 'path': path_found}



#
# Collective Knowledge (universal advice about experiments, bugs, models, 
#                       features, optimizations, adaptation, 
#                       community remarks ...)
#
# See CK LICENSE.txt for licensing details
# See CK COPYRIGHT.txt for copyright details
#
# Developer: Grigori Fursin, Grigori.Fursin@cTuning.org, http://fursin.net
#

cfg={}  # Will be updated by CK (meta description of this module)
work={} # Will be updated by CK (temporal data)
ck=None # Will be updated by CK (initialized CK kernel) 

# Local settings

hextra='<center>\n'
hextra+=' [ <a href="http://cKnowledge.org/ai.html">Project website</a> ], '
hextra+=' [ <a href="http://cKnowledge.org/partners.html">Partners</a> ], '
hextra+=' [ <a href="http://dividiti.blogspot.fr/2017/02/we-received-test-of-time-award-for-our.html">CGO\'17 test of time award for our interdisiplinary R&D</a> ], '
hextra+=' [ <a href="http://cKnowledge.org/ai">Community-driven AI R&D powered by CK</a> ], '
hextra+=' [ <a href="https://github.com/dividiti/ck-caffe">CK-Caffe</a> ], '
hextra+=' [ <a href="https://github.com/ctuning/ck-tensorflow">CK-TensorFlow</a> ], '
hextra+=' [ <a href="https://en.wikipedia.org/wiki/Collective_Knowledge_(software)">Wikipedia</a>, \n'
hextra+='<a href="https://www.researchgate.net/publication/304010295_Collective_Knowledge_Towards_RD_Sustainability">paper 1</a>, \n'
hextra+='<a href="https://arxiv.org/abs/1506.06256">Paper 2</a>, \n'
hextra+='<a href="https://www.youtube.com/watch?v=Q94yWxXUMP0">YouTube CK intro</a> ] \n'
hextra+='</center>\n'
hextra+='\n'

form_name='ck_ai_web_form'
onchange='document.'+form_name+'.submit();'

##############################################################################
# Initialize module

def init(i):
    """

    Input:  {}

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """
    return {'return':0}

##############################################################################
# access available CK-AI self-optimizing functions

def show(i):
    """
    Input:  {
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    import copy

    h=''
    st=''

    h+='<center>\n'
    h+='\n\n<script language="JavaScript">function copyToClipboard (text) {window.prompt ("Copy to clipboard: Ctrl+C, Enter", text);}</script>\n\n' 

#    h+='<h2>Aggregated results from Caffe crowd-benchmarking (time, accuracy, energy, cost, ...)</h2>\n'

    h+=hextra

    # Check host URL prefix and default module/action
    rx=ck.access({'action':'form_url_prefix',
                  'module_uoa':'wfe',
                  'host':i.get('host',''), 
                  'port':i.get('port',''), 
                  'template':i.get('template','')})
    if rx['return']>0: return rx
    url0=rx['url']
    template=rx['template']

    url=url0
    action=i.get('action','')
    muoa=i.get('module_uoa','')

    url+='action=index&module_uoa=wfe&native_action='+action+'&'+'native_module_uoa='+muoa
    url1=url

    # Start form
    r=ck.access({'action':'start_form',
                 'module_uoa':cfg['module_deps']['wfe'],
                 'url':url1,
                 'name':form_name})
    if r['return']>0: return r
    h+=r['html']

    # Check available API is modules
    r=ck.access({'action':'search',
                 'module_uoa':cfg['module_deps']['module'],
                 'tags':'ck-ai-json-web-api'})
    if r['return']>0: return r
    l=r['lst']

    if len(l)==0:
       h='<b>WARNING:</b> No CK modules found with CK AI JSON web API\n'
    else:
       dt=[{'name':'', 'value':''}]

       for x in l:
           v=x['data_uid']

           # Load module info
           r=ck.access({'action':'load',
                        'module_uoa':cfg['module_deps']['module'],
                        'data_uoa':v})
           if r['return']>0: return r

           d=r['dict']

           name=d.get('actions',{}).get('ask_ai_web',{}).get('desc','')

           if name!='':
              dt.append({'name':name, 'value':v})

       # Create selector
       ai=i.get('ai_scenario','')

       ii={'action':'create_selector',
           'module_uoa':cfg['module_deps']['wfe'],
           'data':dt,
           'name':'ai_scenario',
           'onchange':onchange, 
           'skip_sort':'yes',
           'selected_value':ai}
       r=ck.access(ii)
       if r['return']>0: return r
       x=r['html']

       h+='<br><b>Select AI scenario with unified <a href="https://github.com/ctuning/ck/wiki/Unifying-AI-API">CK AI JSON API</a> :</b><br><br>'+x+'<br><br><br>'

       # Render scenario
       if ai!='':
          ii=copy.deepcopy(i)

          ii['action']='ask_ai_web'
          ii['module_uoa']=ai

          ii['widget']='yes'
          ii['prepared_url0']=url0
          ii['prepared_url1']=url1
          ii['prepared_form_name']=form_name

          r=ck.access(ii)
          if r['return']>0: return r

          h+='<div id="ck_box_with_shadow">\n'
          h+='\n\n'+r.get('html','')
          h+='</div>\n'

          st+='\n'+r.get('style','')+'\n'

    h+='<h2>Want to survive Cambrian AI/SW/HW explosion but lost in technological chaos?</h2>\n'
    h+='<img src="http://cKnowledge.org/images/ai-cloud-resize.png"><br><br>\n'

    h+='<b>Join the growing <a href="http://cKnowledge.org/partners.html">consortium</a>\n' 
    h+='using and enhancing <a href="http://cKnowledge.org">Collective Knowledge technology</a> to\n'
    h+=' a) clean up this mess, b) reinvent computer engineering and make it more collaborative, reproducible and reusable,\n'
    h+=' c) develop efficient and reliable computer systems from IoT to supercomputers, \n'
    h+=' d) enable open science via reusable and customizable artifacts,\n'
    h+='  and e) eventually enable and accelerate open AI research!\n'

    return {'return':0, 'html':h, 'style':st}

##############################################################################
# CK-AI dashboard

def browse(i):
    """
    Input:  {
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    i['action']='browser'
    i['cid']=''
    i['module_uoa']=''
    i['template']='ck-ai-basic'

    return ck.access(i)

##############################################################################
# ask AI advice via CK JSON API and CK DNN engines

def ask(i):
    """
    Input:  {
              to - which advice ("predict_compiler_flags", "classify_image")

              (local) - if 'yes', use local optimization rather than public from cKnowledge.org/repo

              If predict_compiler_flags:

                compiler - compiler name
                  See GCC: http://cknowledge.org/repo/web.php?native_action=show&native_module_uoa=program.optimization&scenario=8289e0cf24346aa7
                  See LLVM: http://cknowledge.org/repo/web.php?native_action=show&native_module_uoa=program.optimization&scenario=2aaed4c520956635

                scenario
                cpu_name
                features (MILEPOST feature vector: http://ctuning.org/wiki/index.php/CTools:MilepostGCC:StaticFeatures:MILEPOST_V2.1)

              If classify_image

                dnn_engine (caffe, caffe2, tensorflow)
                image  - file with image
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    import os

    to=i.get('to','')
    if to=='':
       return {'return':1, 'error':'--to is not defined'}

    # Check where to look for optimizations
    er=i.get('exchange_repo','')
    if er=='': er=ck.cfg['default_exchange_repo_uoa']
    esr=''

    aa='show_json'
    if i.get('local','')=='yes': 
       er='local'
       esr=''
       aa='show'

    rr={'return':0}

    if to=='predict_compiler_flags':
       # This is just a demo of MILEPOST project combined with CK-powered collective optimization
       # We plan to considerably improve modeling based on our Collective Mind part II vision paper
       # (bringing community to share optimizations, features, models and enable dynamic adaptation).

       # Check if module exists
       r=ck.access({'action':'find',
                    'module_uoa':cfg['module_deps']['module'],
                    'data_uoa':cfg['module_deps']['milepost']})
       if r['return']>0 and r['return']!=16: return r

       if r['return']==16:
          return {'return':1, 'error':'Please, install CK MILEPOST repository using "ck pull repo:reproduce-milepost-project"'}

       compiler=i.get('compiler','')
       if compiler=='':
          return {'return':1, 'error':'--compiler is not defined'}

       scenario=i.get('scenario','')
       if scenario=='': 
          if compiler.lower().startswith('gcc'):
             scenario=cfg['module_deps']['experiment.tune.compiler.flags.gcc.e']
          elif compiler.lower().startswith('llvm'):
             scenario=cfg['module_deps']['experiment.tune.compiler.flags.llvm.e']
          else:
             return {'return':1, 'error':'scenario is not defined (see "ck search module --tags="program optimization,program-features"'}

       cpu_name=i.get('cpu_name','')
       if cpu_name=='':
          return {'return':1, 'error':'--cpu_name is not defined'}

       features=i.get('features',[])

       xfeatures={}
       for k in i:
           if k.startswith('ft'):
              k1=int(k[2:])
              v1=i[k]
              xfeatures[k1]=v1

       for k in range(0,66):
           if k in xfeatures:
              features.append(xfeatures[k])

       if len(features)==0:
          return {'return':1, 'error':'feature vector is not defined'}

       # Search optimization results
       ii={'action':aa,
           'module_uoa':cfg['module_deps']['program.optimization'],
           'repo_uoa':er,
           'remote_repo_uoa':esr,
           'scenario':scenario,
           '__web_prune__compiler':compiler,
           '__web_prune__cpu_name':cpu_name,
           'skip_html':'yes'}

       r=ck.access(ii)
       if r['return']>0: return r

       results=r['results']
       if len(results)==0:
          return {'return':1, 'error':'optimization results are not found for such configuration'}

       if len(results)>1:
          return {'return':1, 'error':'ambiguity - more then 1 configuration found'}

       # Predict
       muoa=results[0]['module_uoa']
       duoa=results[0]['data_uoa']

       ii={'action':aa,
           'module_uoa':cfg['module_deps']['milepost'],
           'repo_uoa':er,
           'remote_repo_uoa':esr,
           'view_solution_'+muoa+'_'+duoa:'',
           'skip_html':'yes'}

       j=1
       for v in features:
           ii['mft'+str(j)]=v
           j+=1

       rr=ck.access(ii)
       if rr['return']>0: return rr

       popt=rr.get('predicted_opt','')

       if popt=='':
          ck.out('WARNING: could not predict optimization')
       else:
          ck.out('Predicted optimization:')
          ck.out('')
          ck.out(popt)

    ############################################################################################################
    elif to=='classify_image':

       engine=i.get('dnn_engine','')
       if engine=='': engine='caffe'

       image=i.get('image','')
       if image=='' or not os.path.isfile(image):
          return {'return':1, 'error':'image not found'}

       r=ck.convert_file_to_upload_string({'filename':image})
       if r['return']>0: return r

       fcb64=r['file_content_base64']

       # Search optimization results
       ii={'action':aa,
           'module_uoa':cfg['module_deps']['model.image.classification'],
           'repo_uoa':er,
           'remote_repo_uoa':esr,
           'dnn_engine':engine,
           'file_content_base64':fcb64,
           'skip_html':'yes'}

       rr=ck.access(ii)
       if rr['return']>0: return rr

       warning=rr.get('warning','')
       prediction=rr.get('prediction','')

       if warning!='':
          ck.out('WARNING: '+warning)

       if prediction!='':
          ck.out('')
          ck.out('Preciction:')
          ck.out('')
          ck.out(prediction)

    else:
       return {'return':1, 'error':'we do not have such scenario yet ('+to+')'}

    return rr
